"""Implementation of the Digest authentication schema."""

import hashlib
import os
import time
from typing import Dict, Optional
from urllib.parse import urlparse

import requests

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.http import HTTPMethod
from multiauth.entities.main import (
    AuthConfigDigest,
    AuthDigestChallenge,
    AuthHashAlgorithmDigest,
    AuthResponse,
    AuthTech,
)
from multiauth.helpers import hash_calculator
from multiauth.manager import User

# from escape_cli.common.user import USER_MANAGER


def send_401_request(url: str) -> AuthDigestChallenge:
    """Sending a 401 request and parsing it according to RFC 2617."""

    challenge = AuthDigestChallenge(
        {
            'realm': None,
            'domain': None,
            'nonce': None,
            'opaque': None,
            'algorithm': None,
            'qop_options': None,
        },
    )

    # Send an empty get request to get the parameters necessary for the authentication
    response = requests.get(url, timeout=5)

    # We need to parse the response
    if response.headers.get('WWW-Authenticate'):
        parameters_list = response.headers['WWW-Authenticate'].replace(', ', ' ').split(' ')[1::]

        parameters = {}
        for parameter in parameters_list:
            temp = parameter.split('=')
            parameters[temp[0]] = temp[1].replace('"', '')

        challenge['realm'] = parameters.get('realm')

        if parameters.get('domain'):
            challenge['domain'] = parameters.get('domain')

        else:
            parsed_url = urlparse(url)
            challenge['domain'] = parsed_url.path or '/'
            if parsed_url.query and challenge['domain'] is not None:
                challenge['domain'] += '?' + parsed_url.query

        challenge['nonce'] = parameters.get('nonce')
        challenge['opaque'] = parameters.get('opaque')

        if parameters.get('algorithm'):
            challenge['algorithm'] = AuthHashAlgorithmDigest(parameters.get('algorithm'))
        else:
            challenge['algorithm'] = AuthHashAlgorithmDigest.MD5

        challenge['qop_options'] = parameters.get('qop')

    return challenge


# pylint: disable=[too-many-branches, too-many-statements]
def digest_config_parser(schema: Dict) -> AuthConfigDigest:
    """This function parses the Digest schema and checks if all necessary fields exist."""

    auth_config = AuthConfigDigest(
        {
            'url': '',
            'realm': '',
            'nonce': '',
            'algorithm': AuthHashAlgorithmDigest.MD5,
            'domain': '',
            'method': 'POST',
            'qop': None,
            'nonce_count': None,
            'client_nonce': None,
            'opaque': None,
            'headers': None,
        },
    )

    if not schema.get('url'):
        raise AuthenticationError('Please provide a URL to the web application')
    auth_config['url'] = schema['url']

    # Now send a 401 Request to get the needed parameters
    parameters = send_401_request(auth_config['url'])

    # Now fill the optional parameters
    # Hopefully this parsing is compatible with RFC 2617
    # Note all of these are just operations found in RFC 2617

    if schema['options'].get('realm'):
        auth_config['realm'] = schema['options'].get('realm')
    else:
        if parameters['realm'] is None:
            raise AuthenticationError(
                'Cannot retrieve the value of the realm from the server. Please provide the realm value',
            )
        auth_config['realm'] = parameters['realm']

    if schema['options'].get('nonce'):
        auth_config['nonce'] = schema['options'].get('nonce')
    else:
        if parameters['nonce'] is None:
            raise AuthenticationError(
                'Cannot retrieve the value of the nonce from the server. Please provide the value of the nonce',
            )
        auth_config['nonce'] = parameters['nonce']

    if schema['options'].get('algorithm'):
        auth_config['algorithm'] = AuthHashAlgorithmDigest(schema['options'].get('algorithm'))
    else:
        if parameters['algorithm'] is None:
            raise AuthenticationError('No value for parameters algorithm')
        auth_config['algorithm'] = parameters['algorithm']

    if not schema['options'].get('method'):
        raise AuthenticationError('Please provide the used method in the API')

    auth_config['method'] = schema['options'].get('method')

    auth_config['qop'] = schema['options'].get('qop')
    if not auth_config['qop']:
        auth_config['qop'] = parameters['qop_options']

    auth_config['opaque'] = schema['options'].get('opaque')
    if not auth_config['opaque']:
        auth_config['opaque'] = parameters['opaque']

    auth_config['nonce_count'] = schema['options'].get('nonce_count')
    if not auth_config['nonce_count']:
        if auth_config['qop'] is not None and (auth_config['qop'] == 'auth' or 'auth' in auth_config['qop'].split(',')):
            nonce_count = 1
            auth_config['nonce_count'] = f'{nonce_count:08x}'

    auth_config['client_nonce'] = schema['options'].get('client_nonce')
    if not auth_config['client_nonce']:
        if auth_config['qop'] is not None and (auth_config['qop'] == 'auth' or 'auth' in auth_config['qop'].split(',')):
            # Taken from the request library for auth
            s = str(auth_config['nonce_count']).encode('utf-8')
            s += auth_config['nonce'].encode('utf-8')
            s += time.ctime().encode('utf-8')
            s += os.urandom(8)

            auth_config['client_nonce'] = hashlib.sha1(s).hexdigest()[:16]  # noqa: S324

    if parameters['domain'] is not None:
        auth_config['domain'] = parameters['domain']
    else:
        raise AuthenticationError('No value for parameters domain')

    auth_config['headers'] = schema['options'].get('headers')

    return auth_config


def digest_auth_attach(
    user: User,
    auth_config: AuthConfigDigest,
    method: Optional[HTTPMethod],
) -> AuthResponse:
    """This function attaches the user credentials to the schema and generates the proper authentication response."""

    auth_response = AuthResponse(
        {
            'headers': {},
            'tech': AuthTech.DIGEST,
        },
    )

    # Response Calculator
    def kd(secret: str, data: str) -> str:
        return hash_calculator(auth_config['algorithm'], f'{secret}:{data}')

    # First take user credentials
    username, password = user.get_credentials_pair()

    # Now we have to start calculating the response to the challenge
    a1 = f'{username}:{auth_config["realm"]}:{password}'
    if method:
        a2 = f'{method}:{auth_config["domain"]}'
    else:
        a2 = f'{auth_config["method"]}:{auth_config["domain"]}'

    ha1 = hash_calculator(auth_config['algorithm'], a1)
    ha2 = hash_calculator(auth_config['algorithm'], a2)

    if (
        auth_config['algorithm'] == AuthHashAlgorithmDigest.MD5_SESS
        or auth_config['algorithm'] == AuthHashAlgorithmDigest.SHA_256_SESS
        or auth_config['algorithm'] == AuthHashAlgorithmDigest.SHA_512_256_SESS
    ):
        ha1 = hash_calculator(auth_config['algorithm'], f'{ha1}:{auth_config["nonce"]}:{auth_config["client_nonce"]}')

    response = ''

    if auth_config['qop'] is not None and (auth_config['qop'] == 'auth' or 'auth' in auth_config['qop'].split(',')):
        temp = f'{auth_config["nonce"]}:{auth_config["nonce_count"]}:{auth_config["client_nonce"]}:auth:{ha2}'
        response = kd(ha1, temp)

    else:
        response = kd(ha1, f'{auth_config["nonce"]}:{ha2}')

    header_value = f'username="{username}", realm="{auth_config["realm"]}", nonce="{auth_config["nonce"]}", uri="{auth_config["domain"]}, response="{response}"'  # noqa: E501

    if auth_config['opaque']:
        header_value += f', opaque="{auth_config["opaque"]}"'
    if auth_config['algorithm']:
        header_value += f', algorithm="{auth_config["algorithm"].value.upper()}"'
    if auth_config['qop']:
        header_value += f', qop="auth", nc="{auth_config["nonce_count"]}", cnonce="{auth_config["client_nonce"]}"'

    # Add token to the current user
    user.set_token(header_value, None)

    header = {}
    header['Authorization'] = f'Digest {header_value}'

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in header:
                header[name] += ', ' + value

            else:
                header[name] = value

    auth_response['headers'] = header

    return auth_response


def digest_authenticator(
    user: User,
    schema: Dict,
    method: Optional[HTTPMethod],
) -> AuthResponse:
    """This function is a wrapper function that implements the Digest authentication schema.

    It starts by sending an empty get request which allows it to fetch the configuration
    for the digest from the server using the WWW-Authenticate. After that,
    it sends the credentials using the options provided from the server.
    """

    auth_config = digest_config_parser(schema)
    return digest_auth_attach(user, auth_config, method)



"""Implementation of the API Key authentication schema."""

import base64
from typing import Dict

from multiauth.entities.main import AuthResponse, AuthTech
from multiauth.manager import User


def basic_authenticator(
    user: User,
    schema: Dict,
) -> AuthResponse:
    """This function implement the `Basic Authentication` Schema.

    It simply takes the username and password from the current working user, appends the password to the username,
    and base64 encode them. Finally it adds them to the authentication header in the HTTP request
    """

    auth_response = AuthResponse(
        {
            'headers': {},
            'tech': AuthTech.BASIC,
        },
    )

    # Take the username and password from the user in the configuration file
    username, password = user.get_credentials_pair()

    value = username + ':' + password
    # Encode base64 the value
    encoded_value = base64.b64encode(value.encode('ascii'))
    header_value = encoded_value.decode('ascii')

    # Add the token to the current user
    user.set_token(header_value, None)

    auth_response['headers']['Authorization'] = 'Basic ' + header_value

    if 'options' not in schema:
        return auth_response

    # Append the optional headers to the header
    headers = schema['options'].get('headers')
    if headers:
        for name, value in headers.items():
            # Resolving duplicate keys
            if name in auth_response['headers']:
                auth_response['headers'][name] += ', ' + value
            else:
                auth_response['headers'][name] = value

    return auth_response



# pylint: disable=missing-docstring, too-few-public-methods

"""Abstract representation."""

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple

from multiauth.entities.http import HTTPMethod
from multiauth.entities.main import AuthTech, AuthType, JWTToken, Token


class IUser(ABC):

    """The base class for the user entity."""

    _auth_schema: Optional[str]
    _auth_tech: AuthTech
    _auth_type: Optional[AuthType]
    _credentials: Optional[Dict[str, Any]]
    _expired_token: Optional[Token]
    _expires_in: Optional[float]
    _headers: Optional[Dict[str, Any]]
    _refresh_token: Optional[Token]
    _token_info: Optional[JWTToken]
    _token: Optional[Token]

    @abstractmethod
    def reset(self) -> None:
        pass

    @abstractmethod
    def set_token(
        self,
        token: Optional[Token],
        expires_in: Optional[float],
    ) -> None:
        pass

    @abstractmethod
    def get_credentials_pair(self) -> Tuple[str, str]:
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass


class IMultiAuth(ABC):
    _logger: logging.Logger
    _authrc: Optional[str]

    _manager: Any
    _headers: Dict[str, Dict]
    _auths: Dict

    @property
    @abstractmethod
    def headers(self) -> Dict[str, Dict]:
        pass

    @property
    @abstractmethod
    def auths(self) -> Dict:
        pass

    @property
    @abstractmethod
    def users(self) -> Dict:
        pass

    @property
    @abstractmethod
    def schemas(self) -> Dict:
        pass

    @abstractmethod
    def sign(
        self,
        url: str,
        username: str,
        method: HTTPMethod,
        headers: Dict[str, str],
        formatted_payload: Any,
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def reauthenticate(
        self,
        username: str,
        additional_headers: Optional[Dict[str, str]] = None,
        public: bool = False,
    ) -> Tuple[Dict[str, str], Optional[str]]:
        pass

    @abstractmethod
    def authenticate(
        self,
        username: str,
    ) -> Tuple[Dict[str, str], str]:
        pass

    @abstractmethod
    def authenticate_users(self) -> Dict[str, Optional[Token]]:
        pass



"""Generate authrc from -H or cURL."""

import base64
import json
from typing import Dict, List, Optional, Union, cast
from urllib.parse import parse_qs

import graphql

from multiauth.entities.http import HTTPMethod
from multiauth.entities.main import AuthTech, RCFile
from multiauth.utils import setup_logger, uncurl

POTENTIAL_FIELD_NAME = ['token']

LOGGER = setup_logger()


def urlencoded_to_json(data: Optional[str]) -> Optional[str]:
    """This function transforms data in application/x-www-form-urlencoded to json data."""

    if data is None:
        return None

    new_form = parse_qs(data)
    json_data: Dict = {}
    for name, value in new_form.items():
        if len(value) == 1:
            json_data[name] = value[0]
        else:
            json_data[name] = value

    return json.dumps(json_data)


def deserialize_headers(headers: Union[Dict[str, str], List[str], str]) -> Dict[str, str]:
    """Convert headers to standard format."""

    res: Dict[str, str] = {}

    if isinstance(headers, str):
        headers = [headers]

    if isinstance(headers, list):
        for header in headers:
            header_split = header.split(':', 1)
            res[header_split[0].strip()] = header_split[1].strip()

        return res

    return headers


def _manual_fill(headers: Union[Dict[str, str], List[str], str]) -> RCFile:
    """Serialize raw headers in "manual" auth format."""

    headers_dict = deserialize_headers(headers)

    auth_name: str = 'manual_headers'

    return RCFile(
        {
            'auth': {
                auth_name: {
                    'tech': AuthTech.MANUAL.value,
                },
            },
            'users': {'manual_user': {'headers': headers_dict, 'auth': auth_name}},
        },
    )


def _basic_fill(
    headers: Dict[str, str],
    authorization_header: str,
) -> RCFile:
    """Convert basic headers to curl."""

    # Then the type of authentification is basic

    decoded_value: str = cast(str, base64.b64decode(authorization_header.split(' ')[1]))
    username, password = decoded_value.split(':', 1)

    # The JSON schema for every authentication scheme
    rcfile = RCFile(
        {
            'users': {
                'user_basic': {
                    'auth': 'auth_basic',
                    'username': username,
                    'password': password,
                },
            },
            'auth': {'auth_basic': {'tech': AuthTech.BASIC.value}},
        },
    )

    optional_headers: Dict = {}
    for key, value in headers.items():
        if 'authorization' not in key.lower():
            optional_headers[key] = value

    if optional_headers:
        rcfile['auth']['auth_basic']['options'] = {'headers': optional_headers}

    return rcfile


def _rest_fill(
    rest_document: Dict,
    url: str,
    method: HTTPMethod,
    headers: Dict[str, str],
) -> RCFile:
    """This function fills the rest file."""

    # The JSON schema for every authentication scheme
    return RCFile(
        {
            'users': {'user1': {'auth': 'schema1', **rest_document}},
            'auth': {
                'schema1': {
                    'tech': AuthTech.REST.value,
                    'url': url,
                    'method': method,
                    'options': {
                        'headers': headers,
                    },
                },
            },
        },
    )


def _graphql_fill(
    graphql_document: Dict,
    url: str,
    method: HTTPMethod,
    headers: Dict[str, str],
    variables: Optional[Dict] = None,
) -> RCFile:
    """This function fills the graphql escaperc file."""

    variables = variables or {}

    # Now we need to get the user information
    credentials: Dict = {}
    if variables and graphql_document['definitions'][0]['variable_definitions']:
        for variable in graphql_document['definitions'][0]['variable_definitions']:
            variable_name = variable['variable']['name']['value']
            if variable_name in variables:
                credentials[variable_name] = variables[variable_name]

    else:
        arguments = graphql_document['definitions'][0]['selection_set']['selections'][0]['arguments']
        if isinstance(arguments, list):
            for argument in arguments:
                if argument['value'].get('fields') is None:
                    credentials[argument['name']['value']] = argument['value']['value']
                else:
                    credentials[argument['name']['value']] = {}
                    for input_object_field in argument['value']['fields']:
                        credentials[argument['name']['value']][
                            input_object_field['name']['value']
                        ] = input_object_field['value']['value']

    rcfile = RCFile(
        {
            'users': {
                'user1': {
                    'auth': 'schema1',
                    **credentials,
                },
            },
            'auth': {
                'schema1': {
                    'tech': AuthTech.GRAPHQL.value,
                    'url': url,
                    'method': method,
                    'mutation_name': graphql_document['definitions'][0]['selection_set']['selections'][0]['name'][
                        'value'
                    ],
                    'options': {'operation': graphql_document['definitions'][0]['operation']},
                },
            },
        },
    )

    # Now regarding the field
    for field in graphql_document['definitions'][0]['selection_set']['selections'][0]['selection_set']['selections']:
        if field['name']['value'].lower() in POTENTIAL_FIELD_NAME:
            rcfile['auth']['schema1']['mutation_field'] = field['name']['value']
            break

    if headers:
        rcfile['auth']['schema1']['options']['headers'] = headers

    return rcfile


# pylint: disable=too-many-branches, too-many-statements
def curl_to_escaperc(curl: str) -> Optional[RCFile]:
    """This function transforms the curl request to an escaperc file."""

    # First we uncurl
    parsed_content = uncurl(curl)

    # First thing we have to check if in the headers, there is a basic authentication or a token already
    for header_prefix, header_value in parsed_content.headers.items():
        if 'authorization' in header_prefix.lower():
            if 'basic' in header_value.lower():
                LOGGER.info('Type of authetication detected: Basic')
                return _basic_fill(parsed_content.headers, header_value)
            # Here we assume it's manual headers.
            if parsed_content.data is None:
                LOGGER.info('Type of authetication detected: Manual Headers')
                return _manual_fill(parsed_content.headers)

    if parsed_content.data is not None:
        query: Optional[Dict] = None
        try:
            query = json.loads(parsed_content.data)
        except Exception:
            try:  # the request sent is not sent as application/json, let's try application/x-www-form-urlencoded
                json_data = urlencoded_to_json(parsed_content.data)
                if json_data is not None:
                    query = json.loads(json_data)
            except Exception:
                LOGGER.debug('The `data` attribute of the cURL is not JSONable')

        if query is not None:
            if query.get('query') is not None:
                LOGGER.info('Type of authetication detected: GraphQL')
                graphql_tree = graphql.parse(query['query']).to_dict()
                return _graphql_fill(
                    graphql_tree,
                    parsed_content.url,
                    parsed_content.method,
                    parsed_content.headers,
                    query.get('variables'),
                )

            LOGGER.info('Type of authetication detected: REST')
            return _rest_fill(query, parsed_content.url, parsed_content.method, parsed_content.headers)

    LOGGER.info('We could not determine any authentication method from the cURL.')
    return None



"""Implementation of the AWS authentication schema."""

import datetime
import hashlib
import hmac
from copy import deepcopy
from typing import Any, Dict, cast
from urllib.parse import urlparse

import boto3
import jwt
from pycognito.aws_srp import AWSSRP  # type: ignore[import]

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.http import HTTPMethod, Location
from multiauth.entities.main import AuthAWSType, AuthResponse, AuthTech
from multiauth.entities.providers.aws import AuthConfigAWS, AuthHashalgorithmHawkandAWS
from multiauth.helpers import get_secret_hash
from multiauth.manager import User


def aws_check_type(
    user: User,
    schema: Dict,
) -> AuthAWSType:
    """A function that returns the type of the AWS Authentication."""

    auth_config = aws_config_parser(user, schema)

    return auth_config['type']


def aws_config_parser(
    user: User,
    schema: Dict,
) -> AuthConfigAWS:
    """This function parses the Digest schema and checks if all necessary fields exist."""

    auth_config = AuthConfigAWS(
        {
            'type': AuthAWSType.USER_PASSWORD_AUTH,
            'region': '',
            'client_id': None,
            'method': None,
            'service_name': None,
            'hash_algorithm': None,
            'pool_id': None,
            'client_secret': None,
            'location': Location.HEADERS,
            'header_name': None,
            'header_prefix': None,
            'headers': None,
        },
    )

    # Start by taking the type
    if not schema.get('type'):
        raise AuthenticationError('Please provide the type of AWS authentication that you want to perform')
    if not schema.get('region'):
        raise AuthenticationError('Please provide the region in which the service exists (eg: us-east-1)')
    if not schema.get('location'):
        raise AuthenticationError('Please provide with the location where the headers should be added')

    auth_config['type'] = schema['type']
    user.auth_type = schema['type']
    auth_config['region'] = schema['region']
    auth_config['location'] = Location(schema['location'])

    if auth_config['type'] == AuthAWSType.AWS_SIGNATURE:
        if not schema.get('service_name'):
            raise AuthenticationError('Please provide the service name in which you are trying to access (eg: EC2)')
        if not schema.get('method'):
            raise AuthenticationError('Please provide the method of the requests (usually POST)')
        if not schema.get('hash_algorithm'):
            raise AuthenticationError('Please provide the hashing algorithm')

        auth_config['service_name'] = schema['service_name']
        auth_config['method'] = schema['method']
        auth_config['hash_algorithm'] = AuthHashalgorithmHawkandAWS(schema['hash_algorithm'])

    else:
        if not schema.get('client_id'):
            raise AuthenticationError('Please provide the client ID')
        auth_config['client_id'] = schema['client_id']

        if auth_config['type'] == AuthAWSType.USER_SRP_AUTH:
            if not schema.get('pool_id'):
                raise AuthenticationError('Please provide the AWS cognito pool ID')
            auth_config['pool_id'] = schema['pool_id']

    # Options
    if 'options' in schema:
        auth_config['client_secret'] = schema['options'].get('client_secret')
        auth_config['header_name'] = schema['options'].get('header_name')
        auth_config['header_prefix'] = schema['options'].get('header_prefix')
        auth_config['headers'] = schema['options'].get('headers')

    return auth_config


def aws_user_password_handler(
    user: User,
    auth_config: AuthConfigAWS,
) -> Dict:
    """This function is the handler for the USER_PASSWORD_AUTH authentication flow."""

    # First we have to fetch the user credentials from the user
    username, password = user.get_credentials_pair()

    # Now we have to initiate the client
    client = boto3.client('cognito-idp', region_name=auth_config['region'])

    # Now we have to create the parameters
    parameters: Dict[str, str] = {
        'USERNAME': username,
        'PASSWORD': password,
    }

    if auth_config['client_secret'] is not None:
        client_id = cast(str, auth_config['client_id'])
        parameters['SECRET_HASH'] = get_secret_hash(username, client_id, auth_config['client_secret'])

    # Now we have to initiate the connection
    return client.initiate_auth(
        ClientId=auth_config['client_id'],
        AuthFlow=auth_config['type'].value,
        AuthParameters=parameters,
    )


def aws_user_srp_handler(
    user: User,
    auth_config: AuthConfigAWS,
) -> Dict:
    """This function is the handler for the USER_SRP_AUTH authentication flow."""

    # First we have to fetch the user credentials from the user
    username, password = user.get_credentials_pair()

    # Now we have to initiate the client
    client = boto3.client('cognito-idp', region_name=auth_config['region'])

    # Now we to make the connection and get the token
    connection = AWSSRP(
        username=username,
        password=password,
        pool_id=auth_config['pool_id'],
        client_id=auth_config['client_id'],
        client_secret=auth_config['client_secret'],
        client=client,
    )

    return connection.authenticate_user()


def aws_auth_attach(
    user: User,
    auth_config: AuthConfigAWS,
) -> AuthResponse:
    """This function attaches the user credentials to the schema and generates the proper authentication response."""

    aws_response: Dict = {}
    headers: Dict[str, str] = {}

    # First we need to check which authentication flow is used
    if auth_config['type'] == AuthAWSType.USER_PASSWORD_AUTH:
        aws_response = aws_user_password_handler(user, auth_config)
    elif auth_config['type'] == AuthAWSType.USER_SRP_AUTH:
        aws_response = aws_user_srp_handler(user, auth_config)
    elif auth_config['type'] == AuthAWSType.REFRESH_TOKEN:
        if not user.credentials:
            raise AuthenticationError('Configuration file error. Missing credentials')
        if not user.credentials.get('refresh_token'):
            raise AuthenticationError('Please provide the user with refresh token')
        refresh_token = user.credentials['refresh_token']
        return aws_reauthenticator(user, cast(Dict, auth_config), refresh_token, parse=False)
    else:
        return AuthResponse({'tech': AuthTech.AWS, 'headers': {}})

    # Extract the access_token and the refresh token
    access_token: str = aws_response['AuthenticationResult']['AccessToken']
    refresh_token = aws_response['AuthenticationResult']['RefreshToken']

    # Now we to have prepare the header
    if auth_config['header_name'] is not None:
        headers[auth_config['header_name']] = ''
    else:
        headers['Authorization'] = ''

    if auth_config['header_prefix'] is not None:
        headers[next(iter(headers))] += auth_config['header_prefix'] + ' ' + access_token
    else:
        headers[next(iter(headers))] += 'Bearer ' + access_token

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in headers:
                headers[name] += ', ' + value
            else:
                headers[name] = value

    auth_response: AuthResponse = AuthResponse(
        {
            'tech': AuthTech.AWS,
            'headers': headers,
        },
    )

    # Add the token, the refresh token, and the expiry time to the user manager in order
    # to be accessed by other parts of the program
    # First we have to check if the token is a JWT token (It should be)
    try:
        expiry_time = jwt.decode(
            access_token,
            options={
                'verify_signature': False,
                'verify_exp': True,
            },
        ).get('exp')
    except Exception as e:
        raise AuthenticationError('The received token is not a JWT token') from e

    user.set_token(access_token, expiry_time)
    user.refresh_token = refresh_token

    return auth_response


def aws_authenticator(
    user: User,
    schema: Dict,
) -> AuthResponse:
    """This function is a wrapper function that implements the AWS authentication schema.

    The AWS authentication is based on creating a signature based on the access key and the secret key to the API.
    After creating this signature, the signature is appended to a well crafted authentication header
    """

    auth_config = aws_config_parser(user, schema)
    return aws_auth_attach(user, auth_config)


def aws_reauthenticator(
    user: User,
    schema: Dict,
    refresh_token: str,
    parse: bool = True,
) -> AuthResponse:
    """This function is a function that implements the AWS Authentication reauthentication.

    It takes schema and user as input, and it starts tth reauthentication process using the refreash token
    """

    # Reparse the configuration
    if parse:
        auth_config = aws_config_parser(user, schema)
    else:
        auth_config = cast(AuthConfigAWS, schema)

    # Now we have to initiate the reauth
    if auth_config['type'] == AuthAWSType.AWS_SIGNATURE:
        raise AuthenticationError('The AWS Signature is not handled here')

    client = boto3.client('cognito-idp', region_name=auth_config['region'])

    # Now we have to create the parameters
    parameters: Dict[str, str] = {
        'REFRESH_TOKEN': refresh_token,
    }

    if auth_config['client_secret'] is not None:
        if user.credentials:
            if not user.credentials.get('username'):
                raise AuthenticationError('Please provide the username')

            username: str = user.credentials['username']

        client_id = cast(str, auth_config['client_id'])
        parameters['SECRET_HASH'] = get_secret_hash(username, client_id, auth_config['client_secret'])

    # Now we have to initiate the connection
    response = client.initiate_auth(
        ClientId=auth_config['client_id'],
        AuthFlow='REFRESH_TOKEN_AUTH',
        AuthParameters=parameters,
    )

    # Extract the access_token and the refresh token
    access_token: str = response['AuthenticationResult']['AccessToken']
    new_refresh_token: str = response['AuthenticationResult']['RefreshToken']

    try:
        expiry_time = jwt.decode(
            access_token,
            options={
                'verify_signature': False,
                'verify_exp': True,
            },
        ).get('exp')
    except Exception as e:
        raise AuthenticationError('The received token is not a JWT token') from e

    user.set_token(access_token, expiry_time)
    user.refresh_token = new_refresh_token

    # Now we to have prepare the header
    headers: Dict[str, str] = {}

    if not auth_config['header_name'] is not None:
        headers['Authorization'] = ''
    else:
        headers[auth_config['header_name']] = ''

    if auth_config['header_prefix'] is not None:
        headers[next(iter(headers))] += auth_config['header_prefix'] + ' ' + access_token
    else:
        headers[next(iter(headers))] += 'Bearer ' + access_token

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in headers:
                headers[name] += ', ' + value

            else:
                headers[name] = value

    auth_response: AuthResponse = AuthResponse(
        {
            'tech': AuthTech.AWS,
            'headers': headers,
        },
    )

    return auth_response


# pylint: disable=line-too-long, too-many-locals
def aws_signature(
    user: User,
    schema: Dict,
    headers: Dict[str, str],
    method: HTTPMethod,
    payload: Any,
    url: str,
) -> AuthResponse:
    """This function performs the AWS authentication by using the AWS signature v4."""

    auth_config = aws_config_parser(user, schema)

    # This is done according to https://docs.aws.amazon.com/code-samples/latest/catalog/python-signv4-v4-signing-get-post.py.html

    # First we take the user credentials
    if not user.credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')
    if not user.credentials.get('AccessKey'):
        raise AuthenticationError('Please enter an access key')
    if not user.credentials.get('SecretKey'):
        raise AuthenticationError('Please enter a secret key')

    access_key = cast(str, user.credentials.get('AccessKey'))
    secret_key = cast(str, user.credentials.get('SecretKey'))

    # Now we have to create the signing funcitons
    def _sign(key: bytes, msg: str) -> bytes:
        return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

    def _get_signature_key(key: str, date_stamp: str, region_name: str, service_name: str) -> bytes:
        key_date = _sign(('AWS4' + key).encode('utf-8'), date_stamp)
        key_region = _sign(key_date, region_name)
        key_service = _sign(key_region, service_name)
        return _sign(key_service, 'aws4_request')

    # Now we have to define a couple of variables that will help us in the signing
    time = datetime.datetime.utcnow()
    amz_date = time.strftime('%Y%m%dT%H%M%SZ')
    date_stamp = time.strftime('%Y%m%d')

    # Now we have to analyze the URL
    parsed_url = urlparse(url)

    # We want the host
    host = parsed_url.netloc

    # We also want the path
    path = parsed_url.path

    # Add optional header
    _headers: Dict[str, str] = deepcopy(headers)
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in _headers and value not in _headers[name]:
                _headers[name] += ', ' + value
            else:
                _headers[name] = value

    # Now we have to prepare the headers
    signed_header = 'host;x-amz-date'
    canonical_header = 'host: ' + host + '\n' + 'x-amz-date: ' + amz_date
    for header_name, header_value in _headers.items():
        canonical_header += header_name + ': ' + header_value + '\n'
        signed_header += header_name + ';'

    canonical_header = canonical_header[:-1]
    signed_header = signed_header[:-1]

    # Now we have to hash the payload
    payload_hash = hashlib.sha256(payload.encode('utf-8')).hexdigest()

    # Now we have to create the cannonical URL
    if method == 'POST':
        canonical_request = (
            method + '\n' + path + '\n' + '\n' + canonical_header + '\n' + signed_header + '\n' + payload_hash + '\n'
        )
    else:
        canonical_request = (
            method + '\n' + path + '\n' + payload + '\n' + canonical_header + '\n' + signed_header + '\n' + '\n'
        )

    # Now we have to create the strings to sign
    algorithm = 'AWS4-HMAC-SHA256'
    region: str = auth_config['region']
    service: str = cast(str, auth_config['service_name'])
    credential_scope = date_stamp + '/' + region + '/' + service + '/' + 'aws4_request'
    string_to_sign = (
        algorithm
        + '\n'
        + amz_date
        + '\n'
        + credential_scope
        + '\n'
        + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
    )

    # Calculate the signature
    signing_key = _get_signature_key(secret_key, date_stamp, region, service)
    signature = hmac.new(signing_key, (string_to_sign).encode('utf-8'), hashlib.sha256).hexdigest()

    # Creating the header
    authorization_header = (
        algorithm
        + ' '
        + 'Credential='
        + access_key
        + '/'
        + credential_scope
        + ', '
        + 'SignedHeaders='
        + signed_header
        + ', '
        + 'Signature='
        + signature
    )

    # Add header
    _headers['Authorization'] = authorization_header

    return AuthResponse(
        {
            'tech': AuthTech.AWS,
            'headers': _headers,
        },
    )



# pylint: disable=no-name-in-module

"""Helper functions for the authentication process."""

import base64
import hashlib
import hmac
import json
import re
from json.decoder import JSONDecodeError
from typing import Any, Dict, Match, Optional, Tuple, Union, cast

import jwt
import requests

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.main import AuthHashAlgorithmDigest, AuthResponse, AuthTech, JWTToken, Token
from multiauth.entities.providers.oauth import AuthOAuthlocation
from multiauth.utils import dict_nested_get


def extract_token(
    response: requests.Response,
    tech: AuthTech,
    headers: Dict[str, str],
    refresh_token_name: Optional[str] = None,
) -> Tuple[AuthResponse, Optional[str]]:
    """This function takes the response and tries to extract the tokens.

    This function is mainly a helper function to the REST and the GraphQL authenctication schema.
    The goal of the function is to generate the authentication
    response according to the extracted tokens from the repsonse
    """

    def _find_token(token: Any, response: Any) -> Any:
        """This function finds the value of the token."""
        result = response

        if len(token) > 0:
            token_name = token[0]
            res_token_name: Any = dict_nested_get(result, token_name)
            result = _find_token(token[1:], res_token_name)

        return result

    if response.status_code is None or response.status_code >= 400:
        raise AuthenticationError(f'Response returned for authentication has failed: {response.text}')

    try:
        response_dict = json.loads(response.text)
    except JSONDecodeError as e:
        raise AuthenticationError(
            f'{type(e).__name__}: Response returned by authentication server is invalid: {e}',
        ) from e

    headers_to_add: Dict = {}

    if headers is not None:
        for header_name, header_arg in headers.items():
            while '{{' in header_arg and '}}' in header_arg:
                # regex to find the name of the token inside {{token_name}}
                token_name = cast(Match, re.search('{{(.*)}}', header_arg)).group(1)

                # retrived token from the response
                res_token = _find_token(token_name.split('.'), response_dict)

                try:
                    if res_token is None:
                        raise AuthenticationError("The Authentication token wasn't fetched properly.")

                except AssertionError as e:
                    raise AuthenticationError(
                        f"{type(e).__name__}: The Authentication token wasn't fetched properly.",
                    ) from e

                header_arg = header_arg.replace('{{' + token_name + '}}', res_token)  # noqa: PLW2901

            headers_to_add[header_name] = header_arg

    # Here we are going to retrieve the refresh token from the response
    if refresh_token_name is not None:
        refresh_token: str = _find_token(refresh_token_name.split('.'), response_dict)
        return AuthResponse(tech=tech, headers=headers_to_add), refresh_token

    return AuthResponse(tech=tech, headers=headers_to_add), None


def hash_calculator(
    hash_type: AuthHashAlgorithmDigest,
    input_data: Union[str, bytes],
) -> str:
    """This function determines the appropriate hashing function and returns the hashing of the input."""

    if hash_type in (AuthHashAlgorithmDigest.MD5, AuthHashAlgorithmDigest.MD5_SESS):
        if isinstance(input_data, str):
            input_data = input_data.encode('utf-8')
        return hashlib.md5(input_data).hexdigest()  # noqa: S324

    if hash_type in (AuthHashAlgorithmDigest.SHA_256, AuthHashAlgorithmDigest.SHA_256_SESS):
        if isinstance(input_data, str):
            input_data = input_data.encode('utf-8')
        return hashlib.sha256(input_data).hexdigest()

    if hash_type in (AuthHashAlgorithmDigest.SHA_512_256, AuthHashAlgorithmDigest.SHA_512_256_SESS):
        if isinstance(input_data, str):
            input_data = input_data.encode('utf-8')
        return hashlib.sha512(input_data).hexdigest()

    return ''


def token_endpoint_auth_method(auth_location: AuthOAuthlocation) -> str:
    """This function takes the authorization location that is provided in the configuration
    and determines which token endpoint authentication method should be used by the session."""

    if auth_location == AuthOAuthlocation.BODY:
        return 'client_secret_post'
    if auth_location == AuthOAuthlocation.BASIC:
        return 'client_secret_basic'

    return ''  # type: ignore[unreachable]


def get_secret_hash(
    username: str,
    client_id: str,
    client_secret: str,
) -> str:
    """This function calculates the secret hash used in the AWS cognito
    authentication in case the client secret is provided."""

    message = bytearray(username + client_id, 'utf-8')
    hmac_obj = hmac.new(bytearray(client_secret, 'utf-8'), message, hashlib.sha256)
    return base64.standard_b64encode(hmac_obj.digest()).decode('utf-8')


def jwt_token_analyzer(token: Token) -> JWTToken:
    """This function transforms a JWT token into a defined datatype."""

    # First verify the JWT signature
    try:
        _ = jwt.decode(token, options={'verify_signature': False, 'verify_exp': False})
    except Exception as e:
        raise AuthenticationError('The token provided is not a JWT token') from e

    # First of all we need to decrypt the token
    separated_token = token.split('.')
    token_header: str = separated_token[0]
    token_payload: str = separated_token[1]

    header: Dict = json.loads(base64.urlsafe_b64decode(token_header + '=' * (-len(token_header) % 4)))
    payload: Dict = json.loads(base64.urlsafe_b64decode(token_payload + '=' * (-len(token_payload) % 4)))

    return JWTToken(
        {
            'sig': header['alg'],
            'iss': payload.pop('iss', None),
            'sub': payload.pop('sub', None),
            'aud': payload.pop('aud', None),
            'exp': payload.pop('exp', None),
            'nbf': payload.pop('nbf', None),
            'iat': payload.pop('iat', None),
            'jti': payload.pop('jti', None),
            'other': payload,
        },
    )


# def jwt_token_module(token: Token) -> None:
#     """This function takes in a JWT token as an input and analyzes the token
#     and finally returns some alerts regarding the JWT token."""

#     # First verify the JWT signature
#     try:
#         _ = jwt.decode(token, options={'verify_signature': False, 'verify_exp': False})
#     except Exception:
#         raise AuthenticationError('The token provided is not a JWT token')

#     # First divide the token into the header, payload, and signature
#     seperated_token = token.split('.')
#     token_header: str = seperated_token[0]
#     token_payload: str = seperated_token[1]
#     token_signature: str = seperated_token[2]

#     def _decode(string: Token) -> Dict:
#         return json.loads(base64.urlsafe_b64decode(string + '=' * (-len(string) % 4)))

#     def _encode(string: Dict) -> Token:
#         return base64.urlsafe_b64encode(json.dumps(string, separators=(',', ':')).encode()).decode('UTF-8').strip('=')

#     def _check_none_alg(token_header: str, token_payload: str) -> List[Token]:
#         """This function creates tokens with None signature."""

#         algorithms: List[str] = ['none', 'None', 'NONE', 'nOnE']
#         token_header_decoded: Dict = _decode(token_header)
#         result: List[str] = []

#         for algorithm in algorithms:
#             try:
#                 token_header_decoded['alg'] = algorithm
#             except KeyError:
#                 raise AuthenticationError('The header of the JWT token does not contain alg section')

#             result.append(_encode(token_header_decoded) + '.' + token_payload + '.')

#         return result

#     def _check_null_signature(token_header: str, token_payload: str) -> Token:
#         """This function creates a token with just null signature.

#         No changing of signing algorithm
#         """

#         return token_header + '.' + token_payload + '.'

#     def _check_hs_signature(token_header: str, token_payload: str) -> List[Token]:
#         """This function checks if it is possible to use a token with simply an hash signature."""

#         result: List[str] = []

#         # For sha512
#         new_header = _decode(deepcopy(token_header))
#         new_header['alg'] = 'HS512'
#         _new_header = _encode(new_header)
#         content = _new_header + '.' + token_payload
#         signature_hash_512 = base64.urlsafe_b64encode(hmac.new(''.encode(), content.encode(), hashlib.sha512,
#           ).digest()).decode('UTF-8').strip()
#         result.append(content + '.' + signature_hash_512)

#         # For sha256
#         new_header = _decode(deepcopy(token_header))
#         new_header['alg'] = 'HS256'
#         _new_header = _encode(new_header)
#         content = _new_header + '.' + token_payload
#         signature_hash_256 = base64.urlsafe_b64encode(hmac.new(''.encode(), content.encode(), hashlib.sha256,
#           ).digest()).decode('UTF-8').strip()
#         result.append(content + '.' + signature_hash_256)

#         # For sha384
#         new_header = _decode(deepcopy(token_header))
#         new_header['alg'] = 'HS384'
#         _new_header = _encode(new_header)
#         content = _new_header + '.' + token_payload
#         signature_hash_384 = base64.urlsafe_b64encode(hmac.new(''.encode(), content.encode(), hashlib.sha384,
#           ).digest()).decode('UTF-8').strip()
#         result.append(content + '.' + signature_hash_384)

#         return result

#     def _check_rsa_embed(token_header: str, token_payload: str) -> Token:
#         """Check in case the signature that RSA based."""

#         def _get_rsa_key_pair() -> Tuple[Any, Any]:
#             """Generate RSA keys."""

#             # generate private/public key pair
#             key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)

#             # get public key in OpenSSH format
#             public_key = key.public_key().public_bytes(serialization.Encoding.OpenSSH,
#               serialization.PublicFormat.OpenSSH)

#             # get private key in PEM container format
#             pem = key.private_bytes(
#                 encoding=serialization.Encoding.PEM, format=serialization.PrivateFormat.TraditionalOpenSSL,
#                   encryption_algorithm=serialization.NoEncryption()
#             )

#             # decode to printable strings
#             private_key_str = pem.decode('utf-8')
#             public_key_str = public_key.decode('utf-8')

#             return private_key_str, public_key_str

#         priv_key, pub_key = _get_rsa_key_pair()



"""Manage Client auth flow."""

import json
import logging
import os
import time
from copy import deepcopy
from importlib import resources
from typing import Any, Dict, Optional, Tuple

import jsonschema

from multiauth import static
from multiauth.entities.errors import InvalidConfigurationError
from multiauth.entities.http import HTTPMethod
from multiauth.entities.interfaces import IMultiAuth
from multiauth.entities.main import AuthTech, Token
from multiauth.handlers import auth_handler, reauth_handler
from multiauth.manager import User, UserManager
from multiauth.providers.aws import aws_signature
from multiauth.utils import setup_logger


def load_authrc(
    logger: logging.Logger,
    authrc: Optional[str] = None,
) -> Tuple[Dict, Dict]:
    """Load authrc file."""

    filepath = authrc or os.getenv('AUTHRC')
    if not filepath:
        for path in ['.authrc', '.authrc.json']:
            if os.path.exists(path):
                filepath = path
                break
    elif os.path.exists(os.path.expanduser('~/.multiauth/.authrc')):
        filepath = os.path.expanduser('~/.multiauth/.authrc')

    if not filepath:
        raise InvalidConfigurationError('authrc file not found', path='$')

    logger.info(f'loading authrc file: {filepath}')

    with open(filepath, 'r', encoding='utf-8') as f:
        data = json.load(f)

    if 'headers' in data:
        return load_headers(data['headers'])

    if 'auth' not in data:
        raise InvalidConfigurationError('auth section not found', path='$.auth')

    if 'users' not in data:
        raise InvalidConfigurationError('users section not found', path='$.users')

    return data['auth'], data['users']


def load_headers(headers: Dict[str, str]) -> Tuple[Dict, Dict]:
    """Creates a valid user and auth schema from the headers.

    This is used to be able to pass headers easily.
    """
    if not isinstance(headers, dict):
        raise InvalidConfigurationError('headers must be a dict', path='$.headers')
    auth = {'default_schema': {'tech': 'manual'}}
    users = {'default_user': {'auth': 'default_schema', 'headers': headers}}
    return auth, users


class MultiAuth(IMultiAuth):
    """Multiauth manager."""

    def __init__(
        self,
        auths: Optional[Dict] = None,
        users: Optional[Dict] = None,
        authrc: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Initialize the Auth manager."""

        self._logger = logger or setup_logger()
        self._authrc = authrc

        if auths is None or users is None:
            auths, users = load_authrc(self._logger, authrc)

        self.validate(auths, users)

        self._manager: UserManager = UserManager(self.serialize_users(auths, users))
        self._headers: Dict[str, Dict] = {}
        self._auths = auths

    @property
    def headers(self) -> Dict[str, Dict]:
        """Fetch all user's headers."""

        return self._headers

    @property
    def auths(self) -> Dict[str, User]:
        """Fetch all auths methods."""

        return self._auths

    @property
    def users(self) -> Dict[str, User]:
        """Fetch all users of the internal manager."""

        return self._manager.users

    @property
    def schemas(self) -> Dict:
        """Fetch internal schemas."""

        return self._auths

    @staticmethod
    def validate(
        auths: Dict,
        users: Dict,
    ) -> None:
        """Validate the auth schema and users with json schema."""

        # Load the json schema from static
        with resources.open_text(static, 'auth_schema.json') as f:
            json_schema = json.load(f)

        auth_tech_link: Dict[str, str] = {}
        s_users = ', '.join(auth_tech_link.keys())

        for auth_name, auth in auths.items():
            if auth is None or not isinstance(auth, Dict):
                raise InvalidConfigurationError(message='auth is None or is not a Dict', path=f'$.auth.{auth_name}')
            if 'tech' not in auth:
                raise InvalidConfigurationError(message="'tech' is a required property", path=f'$.auth.{auth_name}')
            if auth['tech'] not in json_schema:
                raise InvalidConfigurationError(
                    message=f'\'{auth["tech"]}\' is not a valid auth tech',
                    path=f'$.auth.{auth_name}.tech',
                )
            auth_tech_link[auth_name] = auth['tech']
            try:
                jsonschema.validate(auth, json_schema[auth['tech']]['authSchema'])
            except (jsonschema.ValidationError, jsonschema.SchemaError) as e:
                raise InvalidConfigurationError(
                    message=e.message,
                    path=f'$.auth.{auth_name}' + str(e.json_path)[2:],
                ) from e

        for username, user in users.items():
            if user is None or not isinstance(user, Dict):
                raise InvalidConfigurationError(message='user is None or is not a Dict', path=f'$.users.{username}')
            if 'auth' not in user:
                raise InvalidConfigurationError(
                    message="'auth' is a required property inside a user",
                    path=f'$.users.{username}',
                )
            if user['auth'] not in auth_tech_link:
                raise InvalidConfigurationError(
                    message=f'Auth references user \'{user["auth"]}\' but the only users defined are: {s_users}',
                    path=f'$.users.{username}.auth',
                )
            try:
                jsonschema.validate(user, json_schema[auth_tech_link[user['auth']]]['userSchema'])
            except (jsonschema.ValidationError, jsonschema.SchemaError) as e:
                raise InvalidConfigurationError(message=e.message, path=f'$.users.{username}' + e.json_path[2:]) from e

    @staticmethod
    def serialize_users(
        auths: Dict,
        users: Dict,
    ) -> Dict[str, User]:
        """Serialize raw user to valid config format."""

        users = deepcopy(users)

        for user, user_info in users.items():
            schema = auths[user_info['auth']]

            _user_credientials: Dict[str, Any] = deepcopy(user_info)
            del _user_credientials['auth']

            _user: User = User(
                {
                    'auth_schema': user_info['auth'],
                    'auth_tech': AuthTech.PUBLIC if user_info['auth'] is None else AuthTech(schema['tech']),
                    'credentials': _user_credientials,
                },
            )

            users[user] = _user

        return users

    def sign(
        self,
        url: str,
        username: str,
        method: HTTPMethod,
        headers: Dict[str, str],
        formatted_payload: Any,
    ) -> Dict[str, str]:
        """Sign a payload before sending it.

        This is a mandatory for AWS Signature.
        """

        if self._manager.users[username].auth_type == 'aws_signature':
            user_info: User = self._manager.users[username]
            auth_headers = aws_signature(
                user_info,
                self._auths[user_info.auth_schema],
                headers,
                method,
                formatted_payload,
                url,
            )
            headers.update(auth_headers['headers'])

        return headers

    def authenticate(
        self,
        username: str,
    ) -> Tuple[Dict[str, str], str]:
        """Authenticate the client using the current user."""

        # Reset the user's headers
        self._headers[username] = {}

        user_info: User = self._manager.users[username]

        # Call the auth handler
        self._logger.info(f'Authenticating user: {username}')
        auth_response = auth_handler(self._auths, user_info)
        if auth_response and isinstance(auth_response, Dict):
            self._headers[username] = auth_response['headers']
            self._logger.info(f'Authentication successful for {username}')

        # In case we provided custom headers, we need to merge them with the ones we got from auth_handler
        if user_info.headers:
            self._headers[username].update(user_info.headers)

        return self._headers[username], username

    def authenticate_users(self) -> Dict[str, Optional[Token]]:
        """Authenticate all the users."""

        tokens: Dict[str, Optional[Token]] = {}
        for user, user_info in self._manager.users.items():
            self._logger.info(f'Authenticating users : {user}')

            if user_info.auth_schema:
                self.authenticate(user)

            tokens[user] = self._manager.users[user].token

        return tokens

    def reauthenticate(
        self,
        username: str,
        additional_headers: Optional[Dict[str, str]] = None,
        public: bool = False,
    ) -> Tuple[Dict[str, str], Optional[str]]:
        """Reauthentication of the user in case of token expiry.

        Args:
            username: The username of the user to reauthenticate.
            additional_headers: Additional headers to add to the returned headers.
            public: If True, do not authenticate the user.

        Returns:
            - The headers
            - The username of the user, if authenticated.
        """

        headers = additional_headers or {}
        user_info = self._manager.users[username]
        expiry_time = user_info.expires_in
        refresh_token = user_info.refresh_token

        # If there is no expiry date, no reauthentication is necessary
        # If the expiry date is more then the current time, no reauthentication is necessary
        if expiry_time and expiry_time < time.time():
            self._logger.info('Token is expired')

            user_info.expired_token = user_info.token

            # If this condition is true, we have to reauthenticate the user
            # But before, we have to check if refresh token exists
            if refresh_token:
                auth_response = reauth_handler(
                    self._auths,
                    user_info,
                    refresh_token,
                )

            else:
                auth_response = auth_handler(
                    self._auths,
                    user_info,
                )

            if auth_response and isinstance(auth_response, Dict):
                self._headers[username] = auth_response['headers']
                self._logger.info('Reauthentication Successful')

        if not public:
            headers.update(self._headers.get(username, {}))

        return headers, None if public else username



"""Implementation of the API Key authentication schema."""

from typing import Dict, cast

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.http import Location
from multiauth.entities.main import AuthConfigApiKey, AuthResponse, AuthTech
from multiauth.manager import User


def apikey_config_parser(schema: Dict) -> AuthConfigApiKey:
    """This function parses the API Key schema and checks if all necessary fields exist."""

    auth_config = AuthConfigApiKey(
        {
            'location': Location.HEADERS,
            'header_name': '',
            'header_prefix': None,
            'headers': None,
        },
    )

    if not schema.get('header_name'):
        raise AuthenticationError('Please provide the key of the API Authentication')
    if not schema.get('location'):
        raise AuthenticationError('Please provide the location to where you want to add the API Key')

    auth_config['header_name'] = cast(str, schema.get('header_name'))
    auth_config['location'] = schema['location']

    if 'options' in schema:
        auth_config['header_prefix'] = schema['options'].get('header_prefix', 'Authorization')
        auth_config['headers'] = schema['options'].get('headers')

    return auth_config


def apikey_auth_attach(
    user: User,
    auth_config: AuthConfigApiKey,
) -> AuthResponse:
    """This function attaches the user credentials to the schema and generates the proper authentication response."""

    auth_response = AuthResponse(
        {
            'headers': {},
            'tech': AuthTech.APIKEY,
        },
    )

    # First take the credentials from the user
    if not user.credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')
    if not user.credentials.get('api_key'):
        raise AuthenticationError("Failed to fetch user's API Key")

    api_key: str = user.credentials['api_key']

    # Add the token to the current user
    user.set_token(api_key, None)

    # Implementation with no expression matching in order to work with mypy
    if auth_config['location'] == Location.HEADERS:
        if auth_config['header_prefix'] is not None:
            auth_response['headers'][auth_config['header_name']] = auth_config['header_prefix'] + ' ' + api_key
        else:
            auth_response['headers'][auth_config['header_name']] = api_key

    if auth_config['location'] == Location.URL:
        pass

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in auth_response['headers']:
                auth_response['headers'][name] += ', ' + value

            else:
                auth_response['headers'][name] = value

    return auth_response


def apikey_authenticator(
    user: User,
    schema: Dict,
) -> AuthResponse:
    """This funciton is a wrapper function that implements the API Key authentication schema.

    It simply takes the API key from the user and addes the api key either
    to the header of the HTTP request or as a parameter of the URL
    """

    auth_config = apikey_config_parser(schema)
    return apikey_auth_attach(user, auth_config)



"""Parse a response to extract auth credentials."""

from typing import Dict, Optional

from multiauth.entities.http import HTTPMethod
from multiauth.entities.main import AuthResponse, AuthTech
from multiauth.manager import User
from multiauth.providers import (
    apikey_authenticator,
    aws_authenticator,
    aws_reauthenticator,
    basic_authenticator,
    digest_authenticator,
    graphql_authenticator,
    manual_authenticator,
    oauth_authenticator,
    oauth_reauthenticator,
    rest_authenticator,
)
from multiauth.providers.graphql import graphql_reauthenticator
from multiauth.providers.rest import rest_reauthenticator
from multiauth.providers.webdriver.routine import webdriver_authenticator


# pylint: disable=no-else-return, too-many-return-statements
def auth_handler(
    schemas: Dict,
    user: User,
    method: Optional[HTTPMethod] = None,
) -> Optional[AuthResponse]:
    """Handles the authentication and returns the authentication response.

    This function takes the current working user as input and checks the schema that the user needs.
    After getting the schema, it executes the appropriate
    authenticator according the tech found in the schema.
    """

    # Fetch authentication schema from users
    schema = schemas[user.auth_schema]

    authentication: AuthTech = user.auth_tech
    response: Optional[AuthResponse] = None
    if authentication == AuthTech.APIKEY:
        response = apikey_authenticator(user, schema)

    elif authentication == AuthTech.AWS:
        response = aws_authenticator(user, schema)

    elif authentication == AuthTech.BASIC:
        response = basic_authenticator(user, schema)

    elif authentication == AuthTech.REST:
        response = rest_authenticator(user, schema)

    elif authentication == AuthTech.WEBDRIVER:
        response = webdriver_authenticator(user, schema)

    # The method parameter added is due to the fact the digest uses the method when hashing
    # Although GraphQL apps use POST by default, we use GET in some of our test
    elif authentication == AuthTech.DIGEST:
        response = digest_authenticator(user, schema, method)

    elif authentication == AuthTech.GRAPHQL:
        response = graphql_authenticator(user, schema)

    elif authentication == AuthTech.HAWK:
        pass

    elif authentication == AuthTech.OAUTH:
        response = oauth_authenticator(user, schema)

    elif authentication == AuthTech.MANUAL:
        response = manual_authenticator(user)

    elif authentication == AuthTech.PUBLIC:
        response = None

    return response


def reauth_handler(
    schemas: Dict,
    user: User,
    refresh_token: str,
) -> Optional[AuthResponse]:
    """Handles the reauthentication and returns the new authentication response.

    This function takes the current working user as input and checks the schema that the user needs.
    After getting the schema, it executes the appropriate
    reauthentication according the tech found in the schema.
    """

    schema = schemas[user.auth_schema]
    authentication = user.auth_tech

    # For now we only have reauthentication with OAuth
    if authentication == AuthTech.OAUTH:
        return oauth_reauthenticator(user, schema, refresh_token)

    if authentication == AuthTech.AWS:
        return aws_reauthenticator(user, schema, refresh_token)

    if authentication == AuthTech.REST:
        return rest_reauthenticator(user, schema, refresh_token)

    if authentication == AuthTech.GRAPHQL:
        return graphql_reauthenticator(user, schema, refresh_token)

    if authentication == AuthTech.WEBDRIVER:
        return webdriver_authenticator(user, schema)

    return None



"""Multiauth types related to HTTP protocol."""

import sys
from enum import Enum, unique

if sys.version_info >= (3, 8):
    from typing import Literal  # pylint: disable=no-name-in-module
else:
    from typing_extensions import Literal

HTTPMethod = Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS', 'TRACE', 'CONNECT']


@unique
class Location(str, Enum):

    """The location where the auth data is added to."""

    HEADERS = 'headers'
    URL = 'url'



"""Auth Providers."""

from multiauth.providers.apikey import apikey_authenticator
from multiauth.providers.aws import aws_authenticator, aws_reauthenticator
from multiauth.providers.basic import basic_authenticator
from multiauth.providers.digest import digest_authenticator
from multiauth.providers.graphql import graphql_authenticator
from multiauth.providers.manual import manual_authenticator
from multiauth.providers.oauth import oauth_authenticator, oauth_reauthenticator
from multiauth.providers.rest import rest_authenticator
from multiauth.providers.webdriver import webdriver_authenticator



"""Implementation of the GraphQL authentication schema."""

import re
from typing import Any, Dict, Match, Optional, cast

import jwt
import requests

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.main import AuthResponse, AuthTech
from multiauth.entities.providers.graphql import AuthConfigGraphQL
from multiauth.helpers import extract_token
from multiauth.manager import User

# from escape_cli.common.user import USER_MANAGER


def format_arguments(credentials: Dict) -> str:
    """Generates the arguments for the graphql authentication schema."""

    arguments: str = ''

    for cred_field, cred_value in credentials.items():
        if isinstance(cred_value, Dict):
            arguments += cred_field + ': {' + format_arguments(cred_value) + '},'

        else:
            arguments += cred_field + ': \"' + cred_value + '\",'

    return arguments[:-1]


def generate_authentication_mutation(
    user: User,
    auth_config: AuthConfigGraphQL,
    credentials: Optional[Dict[str, Any]] = None,
    refresh: bool = False,
    refresh_field: bool = True,
) -> Dict:
    """Generate the graphQL query."""

    # Take the credentials from the users
    if credentials is None:
        credentials = user.credentials
        if not credentials:
            raise AuthenticationError('Configuration file error. Missing credentials')

    try:
        # This variable will host the login information part of the mutation
        # example: (login: admin@ecape.tech, password: "p@ssword123")
        arguments: str = '(' + format_arguments(credentials) + ')'

        # Here we start forming the Mutation string
        # Here we have to take the case if refresh is true
        if refresh:
            if auth_config['refresh_mutation_name'] is None:
                raise AuthenticationError('Configuration file error. Missing refresh_mutation_name')

            if refresh_field and auth_config['refresh_mutation_name'] is not None:
                graphql_query = 'mutation {' + auth_config['refresh_mutation_name'] + arguments + '{ \n'
            else:
                if auth_config['refresh_mutation_name'] is not None:
                    graphql_query = 'mutation {' + auth_config['refresh_mutation_name'] + arguments + '\n'
        else:
            graphql_query = 'mutation {' + auth_config['mutation_name'] + arguments + '{ \n'
        graphql_query += auth_config['mutation_field'] + '\n'
        if auth_config['refresh_field_name'] is not None:
            graphql_query += auth_config['refresh_field_name'] + '\n'
        if auth_config['headers'] is not None:
            for header_arg in auth_config['headers'].values():
                if '{{' in header_arg and '}}' in header_arg:
                    graphql_query += cast(Match, re.search('{{(.*)}}', header_arg)).group(1) + '\n'

        if refresh_field:
            graphql_query = graphql_query[:-1] + '}}'
        else:
            graphql_query = graphql_query[:-1] + '}'

        return {
            'http_method': auth_config['method'],
            'graphql_query': graphql_query,
            'graphql_variables': None,
        }

    except KeyError as error:
        raise KeyError(f'The key {error} is missing in the graphql auth config') from error


def graphql_config_parser(schema: Dict) -> AuthConfigGraphQL:
    """This function parses the GraphQL schema and checks if all necessary fields exist."""

    auth_config = AuthConfigGraphQL(
        {
            'url': '',
            'mutation_name': 'str',
            'mutation_field': '',
            'method': 'POST',
            'cookie_auth': False,
            'operation': 'mutation',
            'header_token_name': None,
            'refresh_mutation_name': None,
            'refresh_field_name': None,
            'refresh_field': True,
            'header_name': None,
            'header_prefix': None,
            'headers': None,
        },
    )

    if not schema.get('url'):
        raise AuthenticationError('Please provide with the authentication URL')
    if not schema.get('mutation_name'):
        raise AuthenticationError('Please provide the mutation name for the authentication')
    if not schema.get('mutation_field'):
        raise AuthenticationError('Please provide the mutation field in the authentication response')

    auth_config['url'] = schema['url']
    if not auth_config['url'].startswith('http'):
        auth_config['url'] = 'https://' + auth_config['url']

    auth_config['mutation_name'] = schema['mutation_name']
    auth_config['mutation_field'] = schema['mutation_field']

    # Options
    if 'options' in schema:
        auth_config['refresh_mutation_name'] = schema['options'].get('refresh_mutation_name')
        auth_config['refresh_field_name'] = schema['options'].get('refresh_field_name')
        auth_config['refresh_field'] = schema['options'].get('refresh_field', True)
        auth_config['operation'] = schema['options'].get('operation', 'mutation')
        auth_config['cookie_auth'] = schema['options'].get('cookie_auth', False)
        auth_config['header_name'] = schema['options'].get('header_name')
        auth_config['header_prefix'] = schema['options'].get('header_prefix')
        auth_config['headers'] = schema['options'].get('headers')
        auth_config['method'] = schema['options'].get('method', 'POST')
        auth_config['header_token_name'] = schema['options'].get('header_token_name')

    return auth_config


# pylint: disable=too-many-branches, too-many-statements
def graphql_auth_attach(
    user: User,
    auth_config: AuthConfigGraphQL,
) -> AuthResponse:
    """This function attaches the user credentials to the schema and generates the proper authentication response."""

    # First we have to generate the graphQL query that we need to send
    graphql_query = generate_authentication_mutation(user, auth_config)
    data: Dict[Any, Any]

    # Create the payload
    if not graphql_query['graphql_variables']:
        data = {'query': graphql_query['graphql_query']}

    else:
        data = {'query': graphql_query['graphql_query'], 'variables': graphql_query['graphql_variables']}

    # Now we need to send the request
    response = requests.request(auth_config['method'], auth_config['url'], json=data, timeout=5)

    # If there is a cookie that is fetched, added it to the auth response header
    cookie_header = response.cookies.get_dict()  # type: ignore[no-untyped-call]
    if cookie_header:
        cookie_header = [f'{name}={value}' for name, value in cookie_header.items()]
        cookie_header = ';'.join(cookie_header)
    else:
        if auth_config['cookie_auth']:
            raise AuthenticationError('No cookie found in the response')

        cookie_header = None

    # Prepare the header in order to fetch the token
    # We are creating a header for the token because the helper function '_extract_token' works like that
    headers: Dict[str, str] = {}

    # Now we want to append the authentication headers
    # There are two parts
    # 1- If auth cookie is enabled, then we simply search add the cookie to the auth response and that is it
    # 2- If auth cookie is disables, we continue the authentication process
    if not auth_config['cookie_auth']:
        if auth_config['header_name'] is None:
            headers['Authorization'] = ''
        else:
            headers[auth_config['header_name']] = ''

        if auth_config['header_prefix'] is not None:
            headers[next(iter(headers))] += (
                auth_config['header_prefix'] + ' ' + '{{' + auth_config['mutation_field'] + '}}'
            )
        else:
            headers[next(iter(headers))] += 'Bearer {{' + auth_config['mutation_field'] + '}}'

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in headers:
                headers[name] += ', ' + value
            else:
                headers[name] = value

    # Append the cookie header and check if the authentication type is a cookie authentication or no
    if cookie_header:
        headers['cookie'] = cookie_header
        if auth_config['cookie_auth']:
            return AuthResponse(
                {
                    'tech': AuthTech.GRAPHQL,
                    'headers': headers,
                },
            )

    token: Optional[str] = None
    auth_response: AuthResponse

    # Fetching token from the header is priorized
    # TODO(antoine@escape.tech): Add support of optional headers (Previously inserted in `headers`)
    if auth_config['header_token_name'] is not None:
        token_key = auth_config['header_token_name']
        token = response.headers.get(token_key)
        if token:
            if auth_config['header_prefix']:
                token_key = auth_config['header_prefix'] + ' ' + token

            auth_response = AuthResponse(
                {
                    'tech': AuthTech.REST,
                    'headers': {
                        token_key: token,
                    },
                },
            )

    if not token:
        # Now fetch the token and create the Authentication Response
        auth_response, refresh_token = extract_token(
            response,
            AuthTech.REST,
            headers,
            auth_config['refresh_field_name'],
        )

        token = auth_response['headers'][next(iter(headers))].split(' ')[1]

    # If the token is not a JWT token, don't add expiry time (No way of knowing if the token is expired or no)
    try:
        expiry_time = jwt.decode(
            token,
            options={
                'verify_signature': False,
                'verify_exp': True,
            },
        ).get('exp')
    except Exception:
        return auth_response

    # Add the token and the expiry time to the user manager in order to be accessed by other parts of the program
    user.set_token(token, expiry_time)

    user.refresh_token = refresh_token

    return auth_response


def graphql_authenticator(
    user: User,
    schema: Dict,
) -> AuthResponse:
    """This function is a wrapper function that implements the GraphQL authentication schema.

    It sends a mutation having the credentials of the user as the arguments to the mutations.
    Once it receives the response, it fetches the tokens and creates the authentication response
    """

    auth_config = graphql_config_parser(schema)
    return graphql_auth_attach(user, auth_config)


def graphql_reauthenticator(
    user: User,
    schema: Dict,
    refresh_token: str,
) -> AuthResponse:
    """This function is a wrapper function that implements the GraphQL reauthentication schema.

    It takes the user information, the schema information, and the refresh token and attempts to reauthenticate
    by sending the refresh token to a muatation and receiving back an access token and a refresh token.
    """

    # Reparse the configuration
    auth_config = graphql_config_parser(schema)

    # Now we have to generate the graphQL query that we need to send
    # To do that we have to generate a dictionary
    credentials: Dict = {auth_config['refresh_field_name']: refresh_token}

    # Now we do the same thing we do in the function above
    # First we have to generate the graphQL query that we need to send
    graphql_query = generate_authentication_mutation(
        user,
        auth_config,
        credentials,
        refresh=True,
        refresh_field=auth_config['refresh_field'],
    )
    data: Dict[Any, Any]

    # Create the payload
    if not graphql_query['graphql_variables']:
        data = {'query': graphql_query['graphql_query']}

    else:
        data = {'query': graphql_query['graphql_query'], 'variables': graphql_query['graphql_variables']}

    # Now we need to send the request
    response = requests.request(auth_config['method'], auth_config['url'], json=data, timeout=5)

    # If there is a cookie that is fetched, added it to the auth response header
    cookie_header = response.cookies.get_dict()  # type: ignore[no-untyped-call]
    if cookie_header:
        cookie_header = [f'{name}={value}' for name, value in cookie_header.items()]
        cookie_header = ';'.join(cookie_header)
    else:
        if auth_config['cookie_auth']:
            raise AuthenticationError('No cookie found in the response')

        cookie_header = None

    # Prepare the header in order to fetch the token
    # We are creating a header for the token because the helper function '_extract_token' works like that
    headers: Dict[str, str] = {}

    # Now we want to append the authentication headers
    # There are two parts
    # 1- If auth cookie is enabled, then we simply search add the cookie to the auth response and that is it
    # 2- If auth cookie is disables, we continue the authentication process
    if not auth_config['cookie_auth']:
        if auth_config['header_name'] is None:
            headers['Authorization'] = ''
        else:
            headers[auth_config['header_name']] = ''

        if auth_config['header_prefix'] is not None:
            headers[next(iter(headers))] += (
                auth_config['header_prefix'] + ' ' + '{{' + auth_config['mutation_field'] + '}}'
            )
        else:
            headers[next(iter(headers))] += 'Bearer {{' + auth_config['mutation_field'] + '}}'

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in headers:
                headers[name] += ', ' + value

            else:
                headers[name] = value

    # Append the cookie header and check if the authentication type is a cookie authentication or no
    if cookie_header:
        headers['cookie'] = cookie_header
        if auth_config['cookie_auth']:
            return AuthResponse(
                {
                    'tech': AuthTech.GRAPHQL,
                    'headers': headers,
                },
            )

    # Now fetch the token and create the Authentication Response
    auth_response, refresh_token_result = extract_token(
        response,
        AuthTech.REST,
        headers,
        auth_config['refresh_field_name'],
    )

    token = auth_response['headers'][next(iter(headers))].split(' ')[1]

    # If the token is not a JWT token, don't add expiry time (No way of knowing if the token is expired or no)
    try:
        expiry_time = jwt.decode(
            token,
            options={
                'verify_signature': False,
                'verify_exp': True,
            },
        ).get('exp')
    except Exception:
        return auth_response

    # Add the token and the expiry time to the user manager in order to be accessed by other parts of the program
    user.set_token(token, expiry_time)

    user.refresh_token = refresh_token_result

    return auth_response



# ruff: noqa:T201

"""Multiauth CLI."""

import argparse
from datetime import date

import pkg_resources

from multiauth.main import MultiAuth
from multiauth.utils import setup_logger

__version__ = pkg_resources.get_distribution('py-multiauth').version


# pylint: disable=trailing-whitespace
def cli() -> None:
    """Entry point of the CLI program."""

    print(
        r"""
__________          _____        .__   __  .__   _____          __  .__
\______   \___.__. /     \  __ __|  |_/  |_|__| /  _  \  __ ___/  |_|  |__
 |     ___<   |  |/  \ /  \|  |  \  |\   __\  |/  /_\  \|  |  \   __\  |  \
 |    |    \___  /    Y    \  |  /  |_|  | |  /    |    \  |  /|  | |   Y  \
 |____|    / ____\____|__  /____/|____/__| |__\____|__  /____/ |__| |___|  /
           \/            \/                           \/                 \/
    """,
    )

    print('    Maintainer   https://escape.tech')
    print('    Blog         https://escape.tech/blog')
    print('    Contribute   https://github.com/Escape-Technologies/py-multiauth')
    print('')
    print(f'   (c) 2021 - { date.today().year } Escape Technologies - Version: {__version__}')
    print('\n' * 2)

    logger = setup_logger()

    parser = argparse.ArgumentParser(description='MultiAuth - Multi-Authenticator CLI')
    parser.add_argument(
        'validate',
        help='Validate a configuration file',
    )
    parser.add_argument(
        '-f',
        '--file',
        help='Configuration file to validate',
        required=False,
    )

    args = parser.parse_args()

    if args.validate:
        if not args.file:
            logger.info('No configuration file provided.')

        try:
            auth = MultiAuth(authrc=args.file)
            for user in auth.users:
                auth.authenticate(user)

            logger.info('Configuration file is valid.')

            for user in auth.users:
                logger.info(f'User: {user} | Headers: {auth.headers[user]}')

        except Exception as e:
            logger.error('Invalid configuration file.')
            raise e



from multiauth.entities.providers.webdriver import SeleniumCommand, SeleniumProject, SeleniumTest


def load_selenium_project(data: dict) -> 'SeleniumProject':
    return SeleniumProject(
        # id=data['id'],
        # version=data['version'],
        # name=data['name'],
        # url=data['url'],
        tests=load_selenium_tests(data),
    )


def load_selenium_tests(data: dict) -> list[SeleniumTest]:
    return [load_selenium_test(test) for test in data['tests']]


def load_selenium_test(test: dict) -> SeleniumTest:
    return SeleniumTest(
        id=test['id'],
        name=test['name'],
        commands=load_selenium_commands(test),
    )


def load_selenium_commands(test: dict) -> list[SeleniumCommand]:
    return [load_selenium_command(command) for command in test['commands']]


def load_selenium_command(command: dict) -> SeleniumCommand:
    return SeleniumCommand(
        id=command['id'],
        # comment=command['comment'],
        command=command['command'],
        target=command['target'],
        targets=command['targets'],
        value=command['value'],
    )



from dataclasses import dataclass


@dataclass
class SeleniumCommand:
    id: str
    # comment: str
    command: str
    target: str
    targets: list[list[str]]
    value: str


@dataclass
class SeleniumTest:
    id: str
    name: str
    commands: list[SeleniumCommand]


@dataclass
class SeleniumProject:
    # id: str
    # version: str
    # name: str
    # url: str
    tests: list[SeleniumTest]



"""User manager."""

import time
from typing import Any, Dict, Optional, Tuple, cast

from multiauth.entities.errors import AuthenticationError, ExpiredTokenError
from multiauth.entities.interfaces import IUser
from multiauth.entities.main import AuthTech, AuthType, JWTToken, Token
from multiauth.helpers import jwt_token_analyzer


# pylint: disable=too-many-instance-attributes
class User(IUser):

    """User entity."""

    def __init__(
        self,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Init user."""

        self.reset()

        if kwargs:
            for key, value in kwargs.items():
                if not key.startswith('_'):
                    key = '_' + key  # noqa: PLW2901

                setattr(self, key, value)

        for token in [self.token, self.refresh_token]:
            if not token:
                continue

            serialized_token = jwt_token_analyzer(token)
            if serialized_token.get('exp'):
                self.expires_in = float(cast(str, serialized_token['exp'])) - time.time()
                if self.expires_in < 0:
                    raise ExpiredTokenError('Token expired.')

    def reset(self) -> None:
        """Reset user."""

        self._auth_schema: Optional[str] = None
        self._auth_tech: AuthTech = AuthTech.PUBLIC
        self._auth_type: Optional[AuthType] = None
        self._credentials: Optional[Dict[str, Any]] = None
        self._expired_token: Optional[Token] = None
        self._expires_in: Optional[float] = None
        self._headers: Optional[Dict[str, Any]] = None
        self._refresh_token: Optional[Token] = None
        self._token_info: Optional[JWTToken] = None
        self._token: Optional[Token] = None

    @property
    def auth_schema(self) -> Optional[str]:
        """Get auth schema."""

        return self._auth_schema

    @auth_schema.setter
    def auth_schema(
        self,
        value: Optional[str],
    ) -> None:
        """Set auth schema."""

        self._auth_schema = value

    @property
    def auth_tech(self) -> AuthTech:
        """Get auth tech."""

        return self._auth_tech

    @auth_tech.setter
    def auth_tech(
        self,
        value: AuthTech,
    ) -> None:
        """Set auth tech."""

        self._auth_tech = value

    @property
    def auth_type(self) -> Optional[AuthType]:
        """Get the authentication type."""

        return self._auth_type

    @auth_type.setter
    def auth_type(
        self,
        auth_type: Optional[AuthType],
    ) -> None:
        """Set the authentication type."""

        self._auth_type = auth_type

    @property
    def credentials(self) -> Optional[Dict[str, Any]]:
        """Get credentials."""

        return self._credentials

    @credentials.setter
    def credentials(
        self,
        value: Dict[str, Any],
    ) -> None:
        """Set credentials."""

        self._credentials = value

    @property
    def expired_token(self) -> Optional[Token]:
        """Get the expired token."""

        return self._expired_token

    @expired_token.setter
    def expired_token(
        self,
        token: Optional[Token],
    ) -> None:
        """Set the expired token."""

        self._expired_token = token

    @property
    def expires_in(self) -> Optional[float]:
        """Get the expiration time."""

        return self._expires_in

    @expires_in.setter
    def expires_in(
        self,
        expires_in: Optional[float],
    ) -> None:
        """Set the expiration time."""

        self._expires_in = expires_in

    @property
    def headers(self) -> Optional[Dict[str, Any]]:
        """Get headers."""

        return self._headers

    @property
    def refresh_token(self) -> Optional[Token]:
        """Get the refresh token."""

        return self._refresh_token

    @refresh_token.setter
    def refresh_token(
        self,
        token: Optional[Token],
    ) -> None:
        """Set the refresh token."""

        self._refresh_token = token

    @property
    def token_info(self) -> Optional[JWTToken]:
        """Get the token info."""

        return self._token_info

    @token_info.setter
    def token_info(
        self,
        token: Optional[JWTToken],
    ) -> None:
        """Set the token info."""

        self._token_info = token

    @property
    def token(self) -> Optional[Token]:
        """Get the token."""

        return self._token

    @token.setter
    def token(
        self,
        token: Optional[Token],
    ) -> None:
        """Set the token."""

        self._token = token

    def set_token(
        self,
        token: Optional[Token],
        expires_in: Optional[float],
    ) -> None:
        """Set token."""

        self.token = token
        self.expires_in = expires_in

        if token:
            try:
                self.token_info = jwt_token_analyzer(token)
            except AuthenticationError:
                pass

    def get_credentials_pair(self) -> Tuple[str, str]:
        """Get credentials (RFC AWS & Basic)."""

        if not self.credentials:
            raise AuthenticationError('Missing credentials.')
        if not self.credentials.get('username'):
            raise AuthenticationError('Please provide a username')
        if not self.credentials.get('password'):
            raise AuthenticationError('Please provide a password')

        return self.credentials['username'], self.credentials['password']

    def to_dict(self) -> Dict[str, Any]:
        """Get user as Dict."""

        return {
            'auth_schema': self.auth_schema,
            'auth_tech': self.auth_tech,
            'auth_type': self.auth_type,
            'credentials': self.credentials,
            'token': self.token,
            'refresh_token': self.refresh_token,
            'expires_in': self.expires_in,
            'expired_token': self.expired_token,
            'token_info': self.token_info,
        }


class UserManager:

    """User manager."""

    _users: Dict[str, User]

    def __init__(
        self,
        users: Optional[Dict[str, User]] = None,
    ) -> None:
        """Initialize the User manager."""

        if not users:
            users = {}

        self._users: Dict[str, User] = users

    def reset(self) -> None:
        """Reset the user manager."""

        self._users = {}

    @property
    def users(self) -> Dict[str, User]:
        """Get all users."""

        return self._users



import logging
import time
from datetime import timedelta

import pkg_resources

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.main import AuthResponse, AuthTech, WebdriverConfig
from multiauth.manager import User
from multiauth.providers.webdriver.core import load_selenium_project
from multiauth.providers.webdriver.extractors import extract_token
from multiauth.providers.webdriver.runner import SeleniumTestRunner

logger = logging.getLogger('multiauth.providers.webdriver')


__version__ = pkg_resources.get_distribution('py-multiauth').version


def webdriver_config_parser(schema: dict) -> WebdriverConfig:
    if not schema.get('extract_location'):
        raise AuthenticationError('Please provide the location to where you want to extract the token')

    if not schema.get('extract_regex'):
        raise AuthenticationError('Please provide the regex to extract the token')

    if not schema.get('project'):
        raise AuthenticationError('Please provide the project to run the webdriver tests')

    if not schema.get('project', {}).get('tests'):
        raise AuthenticationError('Please provide the tests to run the webdriver tests')

    options = schema.get('options') or {}

    auth_config = WebdriverConfig(
        extract_location=schema['extract_location'],
        extract_regex=schema['extract_regex'],
        project=load_selenium_project(schema['project']),
        output_format=options.get('output_format') or 'Authorization: Bearer @token@',
        token_lifetime=options.get('token_lifetime'),
        extract_match_index=options.get('extract_match_index'),
    )

    if auth_config.output_format:
        if '@token@' not in auth_config.output_format:
            raise AuthenticationError('Please provide the token placeholder in the output format (`@token`).')

    if len(auth_config.project.tests) > 1:
        logger.warning(f'Found {len(auth_config.project.tests)}, only the first one will be executed')

    return auth_config


def webdriver_authenticator(user: User, schema: dict) -> AuthResponse:
    auth_config = webdriver_config_parser(schema)

    selenium_test = auth_config.project.tests[0]
    logger.info(f'Webdriver authentication using Multiauth {__version__}')
    logger.info(f'Executing test: {selenium_test.name}')

    with SeleniumTestRunner() as r:
        requests: list = r.run(selenium_test)

    logger.info(f'Finished executing Selenium test. Sent `{len(requests)}` requests')

    token = extract_token(
        auth_config.extract_location,
        auth_config.extract_regex,
        auth_config.extract_match_index,
        requests,
    )
    logger.info('Extracted token')

    formatted_token = auth_config.output_format.replace('@token@', token)
    splitted = formatted_token.split(':')
    header_key = splitted[0].strip()
    header_value = ':'.join(splitted[1:]).strip()
    logger.info(f'Formatted header: {header_key}')

    if auth_config.token_lifetime:
        user.expires_in = time.time() + timedelta(seconds=auth_config.token_lifetime).total_seconds()

    return AuthResponse(
        tech=AuthTech.WEBDRIVER,
        headers={
            header_key: header_value,
        },
    )



"""Rest provider."""

import sys
from typing import Dict, Optional

from multiauth.entities.http import HTTPMethod

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


class AuthConfigRest(TypedDict):

    """Authentication Configuration Parameters of the Rest Method."""

    url: str
    method: HTTPMethod
    cookie_auth: bool
    refresh_url: Optional[str]
    refresh_token_name: Optional[str]
    token_name: Optional[str]
    header_name: Optional[str]
    header_prefix: Optional[str]
    headers: Optional[Dict[str, str]]



"""Implementation of the OAuth authentication schema."""

import time
import uuid
from typing import Dict, Tuple, cast

from authlib.integrations.requests_client import OAuth2Session  # type: ignore[import]

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.http import Location
from multiauth.entities.main import AuthResponse, AuthTech
from multiauth.entities.providers.oauth import AuthConfigOAuth, AuthOAuthGrantType, AuthOAuthlocation, AuthOAuthResponse
from multiauth.entities.providers.webdriver import SeleniumCommand, SeleniumTest
from multiauth.helpers import token_endpoint_auth_method
from multiauth.manager import User
from multiauth.providers.webdriver.core import load_selenium_command
from multiauth.providers.webdriver.runner import SeleniumTestRunner


def authentication_portal(
    url: str,
    callback_url: str,
    login_flow: list[SeleniumCommand],
) -> str | None:
    """This function will open up a browser for the user to enter his credentials during OAuth."""

    with SeleniumTestRunner() as runner:
        requests = runner.run(
            SeleniumTest(
                id=str(uuid.uuid4()),
                name='oauth portal',
                commands=[
                    SeleniumCommand(
                        id=str(uuid.uuid4()),
                        command='open',
                        target=url,
                        targets=[],
                        value='',
                    ),
                ]
                + login_flow
                + [
                    SeleniumCommand(
                        id=str(uuid.uuid4()),
                        command='wait',
                        target=f'request_url_contains={callback_url}',
                        targets=[],
                        value='',
                    ),
                ],
            ),
        )

        for req in requests:
            if callback_url in req.url:
                return req.url

        return None


def extract_oauth_token(
    user: User,
    auth_config: AuthConfigOAuth,
    oauth_response: Dict,
) -> AuthResponse:
    """Extract the token, refresh token, and the expiry time from the OAuth access token response."""

    # Initialize the variables
    auth_response = AuthResponse(
        {
            'headers': {},
            'tech': AuthTech.OAUTH,
        },
    )

    response = AuthOAuthResponse(
        {
            'access_token': '',
            'expires_in': None,
            'refresh_token': None,
        },
    )

    if not oauth_response or not isinstance(oauth_response, Dict):
        raise AuthenticationError('Invalid OAuth Response')
    if not oauth_response.get('access_token'):
        raise AuthenticationError('Invalid OAuth Response')

    response['access_token'] = oauth_response['access_token']
    response['refresh_token'] = oauth_response.get('refresh_token')

    # The expire_at field is the amount of seconds to expire. So we need to calculate the UNIX expiry date
    if oauth_response.get('expires_at'):
        response['expires_in'] = int(oauth_response['expires_at']) + time.time()

    # Now check the location to know where should add the token (header or body)
    if auth_config['location'] == Location.HEADERS:
        auth_response['headers']['authorization'] = auth_config['header_prefix'] + ' ' + response['access_token']

    elif auth_config['location'] == Location.URL:
        pass

    # Add the token, the refresh token, and the expiry time
    # to the user manager in order to be accessed by other parts of the program
    user.set_token(response['access_token'], response['expires_in'])
    user.refresh_token = response['refresh_token']

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in auth_response['headers']:
                auth_response['headers'][name] += ', ' + value
            else:
                auth_response['headers'][name] = value

    return auth_response


def auth_code_session(
    user: User,
    auth_config: AuthConfigOAuth,
) -> OAuth2Session:
    """Creates the authentication code seesion."""

    # First we have to fetch the user credentials from the user
    if not user.credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')
    if not user.credentials.get('client_id'):
        raise AuthenticationError('Please provide the user with client ID')
    if not user.credentials.get('client_secret'):
        raise AuthenticationError('Please provide the user with client secret')

    client_id = user.credentials['client_id']
    client_secret = user.credentials['client_secret']

    # Create an OAuth session using the Authlib library functions
    return OAuth2Session(
        client_id,
        client_secret,
        token_endpoint_auth_method=token_endpoint_auth_method(auth_config['auth_location']),
        scope=auth_config['scope'],
        redirect_uri=auth_config['callback_url'],
    )


def auth_code_handler(
    user: User,
    auth_config: AuthConfigOAuth,
) -> Dict:
    """Handles the authentication code OAuth type."""

    # First initiate the OAuth Session
    client = auth_code_session(user, auth_config)

    # Now after we create the client, we have to create the authentication url
    authentication_url, _ = client.create_authorization_url(
        auth_config['authentication_endpoint'],
        state=auth_config['state'],
        code_verifier=auth_config['code_verifier'],
    )

    authorization_response: str | None = None
    # Now we have to pass the authorization URL and the callback URL to the browser in order to fetch what is necessary
    # The if condition is to simply avoid mypy errors
    if auth_config['callback_url'] is not None:
        authorization_response = authentication_portal(
            authentication_url,
            auth_config['callback_url'],
            login_flow=auth_config['login_flow'],
        )
    if not authorization_response:
        raise AuthenticationError('Authentication Error. Please complete the authentication')

    # Now finally, we have to fetch the access token in order to use it for the applicaiton
    return client.fetch_token(auth_config['token_endpoint'], authorization_response=authorization_response)


def implicit_session(
    user: User,
    auth_config: AuthConfigOAuth,
) -> OAuth2Session:
    """Creates the session for implicit authentication."""

    # First as usual we have to fetch the credentials from the user
    if not user.credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')
    if not user.credentials.get('client_id'):
        raise AuthenticationError('Please provide the user with client ID')

    client_id: str = user.credentials['client_id']

    # Create an OAuth session using the Authlib library functions
    return OAuth2Session(
        client_id,
        token_endpoint_auth_method=token_endpoint_auth_method(auth_config['auth_location']),
        scope=auth_config['scope'],
    )


def implicit_handler(
    user: User,
    auth_config: AuthConfigOAuth,
) -> Dict:
    """Handles the implicit authentication OAuth type."""

    # First initiate the OAuth session
    client = implicit_session(user, auth_config)

    authentication_url, _ = client.create_authorization_url(
        auth_config['authentication_endpoint'],
        state=auth_config['state'],
    )

    authorization_response: str | None = None
    # Now we have to pass the authorization URL and the callback URL to the browser in order to fetch what is necessary
    # The if condition is to simply avoid mypy errors
    if auth_config['callback_url'] is not None:
        authorization_response = authentication_portal(
            authentication_url,
            auth_config['callback_url'],
            auth_config['login_flow'],
        )
    if not authorization_response:
        raise AuthenticationError('Authentication Error. Please complete the authentication')

    return client.fetch_token(authorization_response=authorization_response)


def client_cred_session(
    user: User,
    auth_config: AuthConfigOAuth,
) -> OAuth2Session:
    """Creates the session for client credentials authentication."""

    # First we have to fetch the user credentials from the user
    if not user.credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')
    if not user.credentials.get('client_id'):
        raise AuthenticationError('Please provide the user with client ID')
    if not user.credentials.get('client_secret'):
        raise AuthenticationError('Please provide the user with client secret')

    client_id: str = user.credentials['client_id']
    client_secret: str = user.credentials['client_secret']

    # Create an OAuth session using the Authlib library functions
    return OAuth2Session(
        client_id,
        client_secret,
        token_endpoint_auth_method=token_endpoint_auth_method(auth_config['auth_location']),
        scope=auth_config['scope'],
    )


def client_cred_handler(
    user: User,
    auth_config: AuthConfigOAuth,
) -> Dict:
    """Handles the client credentials authentication OAuth type."""

    # First initiate the OAuth session
    client = client_cred_session(user, auth_config)

    return client.fetch_token(auth_config['token_endpoint'])


def password_cred_session(
    user: User,
    auth_config: AuthConfigOAuth,
) -> Tuple[OAuth2Session, str, str]:
    """Creates the session for password credentials authentication."""

    # First we have to fetch the user credentials from the user
    if not user.credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')
    if not user.credentials.get('client_id'):
        raise AuthenticationError('Please provide the user with client ID')
    if not user.credentials.get('client_secret'):
        raise AuthenticationError('Please provide the user with client secret')
    if not user.credentials.get('username'):
        raise AuthenticationError('Please provide the user with username')
    if not user.credentials.get('password'):
        raise AuthenticationError('Please provide the user with password')

    client_id: str = user.credentials['client_id']
    client_secret: str = user.credentials['client_secret']
    client_username: str = user.credentials['username']
    client_password: str = user.credentials['password']

    # Create an OAuth session using the Authlib library functions
    client = OAuth2Session(
        client_id,
        client_secret,
        token_endpoint_auth_method=token_endpoint_auth_method(auth_config['auth_location']),
        scope=auth_config['scope'],
    )

    return client, client_username, client_password


def password_cred_handler(
    user: User,
    auth_config: AuthConfigOAuth,
) -> Dict:
    """Handles the client credentials authentication OAuth type."""

    client, client_username, client_password = password_cred_session(user, auth_config)

    return client.fetch_token(auth_config['token_endpoint'], username=client_username, password=client_password)


def oauth_config_parser(schema: Dict) -> AuthConfigOAuth:
    """This function parses the OAuth schema and checks if all necessary fields exist."""

    auth_config: AuthConfigOAuth = AuthConfigOAuth(
        {
            'grant_type': AuthOAuthGrantType.AUTH_CODE,
            'authentication_endpoint': None,
            'token_endpoint': None,
            'callback_url': None,
            'scope': '',
            'header_prefix': 'Bearer',
            'auth_location': AuthOAuthlocation.BODY,
            'location': Location.HEADERS,
            'state': None,
            'login_flow': [],
            # 'challenge_method': None,
            'code_verifier': None,
            'headers': None,
        },
    )

    if not schema.get('grant_type'):
        raise AuthenticationError('Please provide the grant type')
    auth_config['grant_type'] = AuthOAuthGrantType(schema.get('grant_type'))
    # USER_MANAGER.set_current_user_auth_type(schema.get('grant_type'))

    # Now according to the grant type we will have to check the parser
    if auth_config['grant_type'] in (AuthOAuthGrantType.AUTH_CODE, AuthOAuthGrantType.IMPLICIT):
        if not schema.get('authentication_endpoint'):
            raise AuthenticationError('Please provide an authentication endpoint')
        auth_config['authentication_endpoint'] = schema.get('authentication_endpoint')

    if auth_config['grant_type'] != AuthOAuthGrantType.IMPLICIT:
        if not schema.get('token_endpoint'):
            raise AuthenticationError('Please provide an token endpoint')
        auth_config['token_endpoint'] = schema.get('token_endpoint')

    if auth_config['grant_type'] in (AuthOAuthGrantType.AUTH_CODE, AuthOAuthGrantType.IMPLICIT):
        if not schema.get('callback_url'):
            raise AuthenticationError('Please provide the authenticaiton endpoint')
        auth_config['callback_url'] = schema.get('callback_url')

    auth_config['scope'] = schema.get('scope')

    if schema.get('header_prefix'):
        auth_config['header_prefix'] = schema['header_prefix']

    if schema.get('auth_location'):
        auth_config['auth_location'] = AuthOAuthlocation(schema.get('auth_location'))

    if not schema.get('location'):
        raise AuthenticationError('Please provide the location')
    auth_config['location'] = Location(schema.get('location'))

    if schema.get('options', {}).get('login_flow'):
        auth_config['login_flow'] = [load_selenium_command(ele) for ele in schema['options']['login_flow']]

    # Options
    if 'options' in schema:
        auth_config['state'] = schema['options'].get('state')
        auth_config['code_verifier'] = schema['options'].get('code_verifier')
        auth_config['headers'] = schema['options'].get('headers')

    return auth_config


def oauth_auth_attach(
    user: User,
    auth_config: AuthConfigOAuth,
) -> AuthResponse:
    """This function attaches the user credentials to the schema and generates the proper authentication
    response according to the grant type."""

    # First according every grant type, we will create a handler
    grant_type: AuthOAuthGrantType = auth_config['grant_type']
    oauth_response: Dict = {}

    if grant_type == AuthOAuthGrantType.AUTH_CODE:
        oauth_response = auth_code_handler(user, auth_config)

    elif grant_type == AuthOAuthGrantType.IMPLICIT:
        oauth_response = implicit_handler(user, auth_config)

    elif grant_type == AuthOAuthGrantType.CLIENT_CRED:
        oauth_response = client_cred_handler(user, auth_config)

    elif grant_type == AuthOAuthGrantType.PASSWORD_CRED:
        oauth_response = password_cred_handler(user, auth_config)

    elif grant_type == AuthOAuthGrantType.REFRESH_TOKEN:
        if not user.credentials:
            raise AuthenticationError('Configuration file error. Missing credentials')
        if not user.credentials.get('refresh_token'):
            raise AuthenticationError('Please provide the user with refresh token')
        refresh_token = user.credentials['refresh_token']
        return oauth_reauthenticator(user, cast(Dict, auth_config), refresh_token, parse=False)

    return extract_oauth_token(user, auth_config, oauth_response)


def oauth_authenticator(
    user: User,
    schema: Dict,
) -> AuthResponse:
    """This function is a wrapper function that implements the OAuth authentication schema.

    It starts by identifying the grant type and then use the appropriate grant type funtion in order
    to authenticate the user to the application.
    """

    auth_config = oauth_config_parser(schema)
    return oauth_auth_attach(user, auth_config)


def oauth_reauthenticator(
    user: User,
    schema: Dict,
    refresh_token: str,
    parse: bool = True,
) -> AuthResponse:
    """This function is a function that implements the OAuth reauthentication.

    It takes the schema and user, and it starts the reauthentication process using the refresh token.
    """

    # Reparse the configuration
    if parse:
        auth_config = oauth_config_parser(schema)
    else:
        auth_config = cast(AuthConfigOAuth, schema)

    # Since authentication requires the existance of an authentication token, Only check the grant type that require
    # an authentication token endpoint as input

    if auth_config['grant_type'] in (AuthOAuthGrantType.AUTH_CODE, AuthOAuthGrantType.REFRESH_TOKEN):
        client = auth_code_session(user, auth_config)

    elif auth_config['grant_type'] == AuthOAuthGrantType.CLIENT_CRED:
        client = client_cred_session(user, auth_config)

    elif auth_config['grant_type'] == AuthOAuthGrantType.PASSWORD_CRED:
        client, _, _ = password_cred_session(user, auth_config)

    elif auth_config['grant_type'] == AuthOAuthGrantType.IMPLICIT:
        client = implicit_session(user, auth_config)

    if auth_config['token_endpoint']:
        new_token = client.refresh_token(auth_config['token_endpoint'], refresh_token)

        return extract_oauth_token(user, auth_config, new_token)

    if auth_config['grant_type'] == AuthOAuthGrantType.REFRESH_TOKEN and not auth_config['token_endpoint']:
        raise AuthenticationError('Please provide the token endpoint')

    return oauth_auth_attach(user, auth_config)



import logging
import re
import time

from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from seleniumwire import webdriver  # type: ignore[import]

from multiauth.entities.errors import AuthenticationError
from multiauth.providers.webdriver.core import SeleniumCommand
from multiauth.providers.webdriver.transformers import target_to_selector_value, target_to_value


class SeleniumCommandHandler:
    driver: webdriver.Firefox
    logger: logging.Logger

    wait_for_seconds: int
    pooling_interval: float

    def __init__(self, driver: webdriver.Firefox) -> None:
        self.driver = driver
        self.logger = logging.getLogger('multiauth.providers.webdriver.seleniumcommandhandler')
        self.wait_for_seconds = 30
        self.pooling_interval = 0.5

    def find_element(self, selector: str, value: str) -> WebElement:
        wait = WebDriverWait(self.driver, self.wait_for_seconds)
        return wait.until(EC.presence_of_element_located((selector, value)))

    def safe_click(self, selector: str, value: str) -> None:
        for _ in range(10):
            try:
                return self.find_element(selector, value).click()
            except Exception as e:
                time.sleep(1)
                last_error = e

        raise AuthenticationError(f'Failed to click element after 10 retries: `{selector}`') from last_error

    def safe_switch_to_frame(self, selector: int | str) -> None:
        for _ in range(10):
            try:
                return self.driver.switch_to.frame(selector)
            except Exception as e:
                time.sleep(1)
                last_error = e

        raise AuthenticationError(f'Failed to click element after 10 retries: `{selector}`') from last_error

    def open(self, command: SeleniumCommand) -> None:
        self.driver.get(command.target)

    def set_window_size(self, command: SeleniumCommand) -> None:
        width, height = command.target.split('x')
        self.driver.set_window_size(int(width), int(height))

    def click(self, command: SeleniumCommand, retries: int | None = None) -> None:
        last_exc = None
        for target_pair in command.targets:
            try:
                selector, value = target_to_selector_value(target_pair)
                return self.safe_click(selector, value)
            except Exception as e:
                logging.info(
                    'Failed to execute click `%s`.`%s`: %s',
                    command.id,
                    target_pair,
                    e,
                )
                last_exc = e

        if retries is None:
            self.logger.info('Retrying click `%s`', command.id)
            self.driver.implicitly_wait(10)
            return self.click(command, retries=1)

        if last_exc:
            raise last_exc
        return None

    def select_frame(self, command: SeleniumCommand) -> None:
        target = command.target

        try:
            if target.startswith('index='):
                index = int(target_to_value(target))
                self.safe_switch_to_frame(index)

            # Check if target is a name or ID
            elif '=' not in target:  # Assumes no "=" character in frame names or IDs
                self.safe_switch_to_frame(target)

            else:
                raise ValueError(f'Unhandled selector type for selectFrame: {target}')
        except Exception as e:
            logging.info(
                'Failed to execute type `%s`.`%s`: %s',
                command.id,
                target,
                e,
            )
            raise e

    def type(self, command: SeleniumCommand) -> None:
        last_exc = None
        for target_pair in command.targets:
            try:
                selector, value = target_to_selector_value(target_pair)
                return self.find_element(selector, value).send_keys(command.value)
            except Exception as e:
                logging.info(
                    'Failed to execute type `%s`.`%s`: %s',
                    command.id,
                    target_pair,
                    e,
                )
                last_exc = e

        if last_exc:
            raise last_exc
        return None

    def mouse_over(self, command: SeleniumCommand) -> None:
        last_exc = None
        for target_pair in command.targets:
            try:
                selector, value = target_to_selector_value(target_pair)
                return (
                    ActionChains(self.driver)
                    .move_to_element(  # type: ignore[no-untyped-call]
                        self.find_element(selector, value),
                    )
                    .perform()
                )
            except Exception as e:
                logging.info(
                    'Failed to execute mouesOver `%s`.`%s`: %s',
                    command.id,
                    target_pair,
                    e,
                )
                last_exc = e

        if last_exc:
            raise last_exc
        return None

    def mouse_out(self, command: SeleniumCommand) -> None:
        last_exc = None
        for target_pair in command.targets:
            try:
                selector, value = target_to_selector_value(target_pair)

                return (
                    ActionChains(self.driver)
                    .move_to_element(  # type: ignore[no-untyped-call]
                        self.find_element(selector, value),
                    )
                    .perform()
                )
            except Exception as e:
                logging.info(
                    'Failed to execute mouseOut `%s`.`%s`: %s',
                    command.id,
                    target_pair,
                    e,
                )
                last_exc = e

        if last_exc:
            raise last_exc
        return None

    def wait(self, command: SeleniumCommand) -> None:
        if command.target:
            cmd, value = command.target.split('=')
            if cmd == 'request_url_contains':
                return self.wait_for_request_url_contains(value)
            return None

        time.sleep(int(command.value))
        return None

    def wait_for_request_url_contains(self, regex: str) -> None:
        started_at = time.time()
        while started_at + self.wait_for_seconds > time.time():
            for request in self.driver.requests:
                if re.match(regex, request.url):
                    return

            time.sleep(self.pooling_interval)



"""Oauth provider."""

import sys
from enum import Enum, unique
from typing import Dict, Optional

from multiauth.entities.http import Location
from multiauth.entities.providers.webdriver import SeleniumCommand

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


@unique
class AuthOAuthlocation(str, Enum):

    """Where the credentials during the OAuth will be sent."""

    BASIC = 'basic'
    BODY = 'body'


@unique
class AuthOAuthGrantType(str, Enum):

    """The grant types of the OAuth."""

    AUTH_CODE = 'auth_code'
    CLIENT_CRED = 'client_cred'
    IMPLICIT = 'implicit'
    PASSWORD_CRED = 'password_cred'
    REFRESH_TOKEN = 'refresh_token'


# class AuthHashAlgorithmOAuth(Enum):

#     """The Available Hashing algorithm for OAuth authentication."""
#     PLAIN = 'plain'
#     SHA_256 = 'sha-256'


class AuthOAuthResponse(TypedDict):

    """The format of the OAuth access token response according to the official documentation."""

    access_token: str
    expires_in: Optional[float]
    refresh_token: Optional[str]


class AuthConfigOAuth(TypedDict):

    """Authentication Configuration Parameters of the OAuth Method."""

    grant_type: AuthOAuthGrantType
    authentication_endpoint: Optional[str]
    token_endpoint: Optional[str]
    callback_url: Optional[str]
    scope: Optional[str]
    header_prefix: str
    auth_location: AuthOAuthlocation
    location: Location
    state: Optional[str]
    login_flow: list[SeleniumCommand]
    # challenge_method: Optional[AuthHashAlgorithmOAuth]
    code_verifier: Optional[str]
    headers: Optional[Dict[str, str]]



import logging
import os
from types import TracebackType
from typing import Any, Self

from selenium.webdriver import firefox
from seleniumwire import webdriver  # type: ignore[import]

from multiauth.providers.webdriver.command import SeleniumCommandHandler
from multiauth.providers.webdriver.core import SeleniumTest


class SeleniumTestRunner:
    driver: webdriver.Firefox | None
    logger: logging.Logger

    def __init__(self) -> None:
        self.driver = None
        self.logger = logging.getLogger('multiauth.providers.webdriver.seleniumtestrunner')

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if self.driver:
            self.driver.quit()
            self.driver = None

    def run(self, test: SeleniumTest) -> Any:
        self.logger.info('Setting up driver..')
        self.driver = self.setup_driver()
        self.logger.info('Initialized firefox browser..')

        hdlr = SeleniumCommandHandler(self.driver)
        cmd_mapping: dict = {
            'open': hdlr.open,
            'setWindowSize': hdlr.set_window_size,
            'click': hdlr.click,
            'type': hdlr.type,
            'mouseOver': hdlr.mouse_over,
            'mouseOut': hdlr.mouse_out,
            'wait': hdlr.wait,
            'selectFrame': hdlr.select_frame,
        }

        for command in test.commands:
            self.logger.info('Executing command: %s (%s)', command.command, command.id)

            if command.command == 'close':
                break

            if command.command not in cmd_mapping:
                raise ValueError(f'Invalid command `{command.command}`')

            try:
                cmd_mapping[command.command](command)
            except Exception as e:
                # self.driver.save_screenshot(f'{command.id}.png')
                raise RuntimeError(f'Failed to execute command `{command.id}`') from e

        return self.driver.requests

    def setup_driver(self) -> webdriver.Firefox:
        firefox_options = firefox.options.Options()
        firefox_options.add_argument('--no-sandbox')
        firefox_options.add_argument('--headless')
        firefox_options.add_argument('--disable-gpu')
        firefox_options.set_preference('browser.download.folderList', 2)
        firefox_options.set_preference('browser.download.manager.showWhenStarting', False)
        firefox_options.set_preference('browser.download.dir', os.getcwd())
        firefox_options.set_preference('browser.helperApps.neverAsk.saveToDisk', 'text/csv')

        driver = webdriver.Firefox(options=firefox_options)

        if proxy := os.getenv('ALL_PROXY'):
            driver.proxy = {
                'https': proxy,
            }

        self.logger.info('Prepared firefox profile..')

        return driver



from selenium.webdriver.common.by import By


def target_to_selector(target: str) -> str:
    if target == 'name':
        return By.NAME
    if target == 'css:finder':
        return By.CSS_SELECTOR
    if target.startswith('xpath'):
        return By.XPATH

    raise Exception(f'Invalid target `{target}`')


def target_to_value(target: str) -> str:
    return target.split('=')[1]


def target_to_selector_value(target_pair: list[str]) -> tuple[str, str]:
    return target_to_selector(target_pair[1]), target_to_value(target_pair[0])



"""Graphql provider."""

import sys
from typing import Dict, Optional

from multiauth.entities.http import HTTPMethod

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import Literal, TypedDict

Operation = Literal['query', 'mutation', 'subscription']


class AuthConfigGraphQL(TypedDict):

    """Authentication Configuration Parameters of the GraphQL Method."""

    url: str
    mutation_name: str
    cookie_auth: bool
    method: HTTPMethod
    mutation_field: str
    operation: Operation
    header_token_name: Optional[str]
    refresh_mutation_name: Optional[str]
    refresh_field_name: Optional[str]
    refresh_field: bool
    header_name: Optional[str]
    header_prefix: Optional[str]
    headers: Optional[Dict[str, str]]



"""Implementation of the Rest authentication schema."""

from typing import Dict, Optional, cast

import jwt
import requests

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.main import AuthResponse, AuthTech
from multiauth.entities.providers.rest import AuthConfigRest
from multiauth.helpers import extract_token
from multiauth.manager import User

# from escape_cli.common.user import USER_MANAGER


def rest_config_parser(schema: Dict) -> AuthConfigRest:
    """This function parses the Rest schema and checks if all the necessary fields exist."""

    auth_config = AuthConfigRest(
        {
            'url': '',
            'method': 'POST',
            'token_name': None,
            'cookie_auth': False,
            'refresh_url': None,
            'refresh_token_name': None,
            'header_name': None,
            'header_prefix': None,
            'headers': None,
        },
    )

    if not schema.get('url'):
        raise AuthenticationError('Please provide the authentication URL')
    if not schema.get('method'):
        raise AuthenticationError('Please provide the HTTP method to use for authentication')

    auth_config['url'] = schema['url']
    auth_config['method'] = schema['method']

    # Options:
    if 'options' in schema:
        auth_config['cookie_auth'] = schema['options'].get('cookie_auth', False)
        if not auth_config['cookie_auth'] and not schema['options'].get('token_name'):
            raise AuthenticationError('Please provide the token name in the authentication response')

        auth_config['refresh_url'] = schema['options'].get('refresh_url')
        auth_config['refresh_token_name'] = schema['options'].get('refresh_token_name')
        auth_config['token_name'] = schema['options'].get('token_name')
        auth_config['header_name'] = schema['options'].get('header_name')
        auth_config['header_prefix'] = schema['options'].get('header_prefix')
        auth_config['headers'] = schema['options'].get('headers')

    return auth_config


# pylint: disable=too-many-branches
def rest_auth_attach(
    user: User,
    auth_config: AuthConfigRest,
) -> AuthResponse:
    """This function attaches the user credentials to the schema and generates the proper authentication response."""

    # First we have to take the credentials from the currently working user
    credentials = user.credentials
    if not credentials:
        raise AuthenticationError('Configuration file error. Missing credentials')

    # Now we need to send the request
    response = requests.request(auth_config['method'], auth_config['url'], json=credentials, timeout=5)

    # If there is a cookie that is fetched, added it to the auth response header
    cookie_header = response.cookies.get_dict()  # type: ignore[no-untyped-call]
    if cookie_header:
        cookie_header = [f'{name}={value}' for name, value in cookie_header.items()]
        cookie_header = ';'.join(cookie_header)
        if auth_config['cookie_auth'] and not cookie_header:
            raise AuthenticationError('Authentication Failed: No cookie was found')
    else:
        cookie_header = None

    # Prepare the header in order to fetch the token
    # We are creating a header for the token because the helper function '_extract_token' works like that
    headers: Dict[str, str] = {}

    # Now we want to append the authentication headers
    # There are two parts
    # 1- If auth cookie is enabled, then we simply search add the cookie to the auth response and that is it
    # 2- If auth cookie is disables, we continue the authentication process
    if not auth_config['cookie_auth']:
        token_name = cast(str, auth_config['token_name'])
        if auth_config['header_name'] is None:
            headers['Authorization'] = ''
        else:
            headers[auth_config['header_name']] = ''

        if auth_config['header_prefix'] is not None:
            headers[next(iter(headers))] += auth_config['header_prefix'] + ' ' + '{{' + token_name + '}}'
        else:
            headers[next(iter(headers))] += 'Bearer {{' + token_name + '}}'

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in headers:
                headers[name] += ', ' + value

            else:
                headers[name] = value

    # Append the cookie header and check if the authentication type is a cookie authentication or no
    if cookie_header:
        headers['cookie'] = cookie_header
        if auth_config['cookie_auth']:
            return AuthResponse(
                {
                    'tech': AuthTech.REST,
                    'headers': headers,
                },
            )

    auth_response, refresh_token = extract_token(response, AuthTech.REST, headers, auth_config['refresh_token_name'])

    token = auth_response['headers'][next(iter(headers))].split(' ')[1]

    # Add the token and the expiry time to the user manager in order to be accessed by other parts of the program
    expiry_time: Optional[float] = None
    try:
        expiry_time = jwt.decode(
            token,
            options={
                'verify_signature': False,
                'verify_exp': True,
            },
        ).get('exp')
    except jwt.exceptions.DecodeError:
        pass
    finally:
        user.set_token(token, expiry_time)

    user.refresh_token = refresh_token

    return auth_response


def rest_authenticator(
    user: User,
    schema: Dict,
) -> AuthResponse:
    """This funciton is a wrapper function that implements the Rest authentication schema.

    It takes the credentials of the user and authenticates them on the authentication URL.
    After authenticating, it fetches the token and adds the token to the
    headers along with optional headers in case the user provided them.
    """

    auth_config = rest_config_parser(schema)
    return rest_auth_attach(user, auth_config)


def rest_reauthenticator(
    user: User,
    schema: Dict,
    refresh_token: str,
) -> AuthResponse:
    """This funciton is a wrapper function that implements the Rest reauthentication schema.

    It takes the user information, the schema information and the refresh token
    and attempts reauthenticating the user using the refresh token
    """

    # Reparse the configuration
    auth_config = rest_config_parser(schema)

    # Now we will do the same thing we are doing in the authentication function
    # First we have to create a payload
    if auth_config['refresh_token_name'] is None or auth_config['refresh_url'] is None:
        raise AuthenticationError('Refresh Token found, please provide the refresh token name and the refresh URL')
    payload: Dict = {auth_config['refresh_token_name']: refresh_token}

    # Now we have to send the payload
    response = requests.request(auth_config['method'], auth_config['refresh_url'], json=payload, timeout=5)

    # If there is a cookie that is fetched, added it to the auth response header
    cookie_header = response.cookies.get_dict()  # type: ignore[no-untyped-call]
    if cookie_header:
        cookie_header = [f'{name}={value}' for name, value in cookie_header.items()]
        cookie_header = ';'.join(cookie_header)
        if auth_config['cookie_auth'] and not cookie_header:
            raise AuthenticationError('Authentication Failed: No cookie was found')
    else:
        cookie_header = None

    # Prepare the header in order to fetch the token
    # We are creating a header for the token because the helper function '_extract_token' works like that
    headers: Dict[str, str] = {}

    # Now we want to append the authentication headers
    # There are two parts
    # 1- If auth cookie is enabled, then we simply search add the cookie to the auth response and that is it
    # 2- If auth cookie is disables, we continue the authentication process
    if not auth_config['cookie_auth']:
        token_name = cast(str, auth_config['token_name'])
        if auth_config['header_name'] is None:
            headers['Authorization'] = ''
        else:
            headers[auth_config['header_name']] = ''

        if auth_config['header_prefix'] is not None:
            headers[next(iter(headers))] += auth_config['header_prefix'] + ' ' + '{{' + token_name + '}}'
        else:
            headers[next(iter(headers))] += 'Bearer {{' + token_name + '}}'

    # Append the optional headers to the header
    if auth_config['headers'] is not None:
        for name, value in auth_config['headers'].items():
            # Resolving duplicate keys
            if name in headers:
                headers[name] += ', ' + value

            else:
                headers[name] = value

    # Append the cookie header and check if the authentication type is a cookie authentication or no
    if cookie_header:
        headers['cookie'] = cookie_header
        if auth_config['cookie_auth']:
            return AuthResponse(
                {
                    'tech': AuthTech.REST,
                    'headers': headers,
                },
            )

    auth_response, refresh_token_result = extract_token(
        response,
        AuthTech.REST,
        headers,
        auth_config['refresh_token_name'],
    )

    token = auth_response['headers'][next(iter(headers))].split(' ')[1]

    # Add the token and the expiry time to the user manager in order to be accessed by other parts of the program
    expiry_time: Optional[float] = None
    try:
        expiry_time = jwt.decode(
            token,
            options={
                'verify_signature': False,
                'verify_exp': True,
            },
        ).get('exp')
    except jwt.exceptions.DecodeError:
        pass
    finally:
        user.set_token(token, expiry_time)

    user.refresh_token = refresh_token_result

    return auth_response



"""Custom types of authentication module."""

import sys
from enum import Enum, unique
from typing import Any, Dict, Optional, Union

from attr import dataclass

from multiauth.entities.http import HTTPMethod, Location
from multiauth.entities.providers.aws import AuthAWSType
from multiauth.entities.providers.oauth import AuthOAuthGrantType
from multiauth.entities.providers.webdriver import SeleniumProject

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


@unique
class AuthHashAlgorithmDigest(str, Enum):

    """The Available Hashing algorithms for Digest Authentication."""

    MD5 = 'md5'
    MD5_SESS = 'md5-sess'
    SHA_256 = 'sha-256'
    SHA_256_SESS = 'sha-256-sess'
    SHA_512_256 = 'sha-512-256'
    SHA_512_256_SESS = 'sha-512-256-sess'


# The Authentication Schemas can be found below
@unique
class AuthTech(str, Enum):

    """Authentication Method Enumeration."""

    APIKEY = 'api_key'
    AWS = 'aws'
    BASIC = 'basic'
    REST = 'rest'
    DIGEST = 'digest'
    GRAPHQL = 'graphql'
    HAWK = 'hawk'
    MANUAL = 'manual'
    PUBLIC = 'public'
    OAUTH = 'oauth'
    WEBDRIVER = 'webdriver'


class AuthDigestChallenge(TypedDict):

    """The format of the challenge in a digest authentication schema as specified by the RFC 2617."""

    realm: Optional[str]
    domain: Optional[str]
    nonce: Optional[str]
    opaque: Optional[str]
    algorithm: Optional[AuthHashAlgorithmDigest]
    qop_options: Optional[str]


class AuthConfigApiKey(TypedDict):

    """Authentication Configuration Parameters of the Api Key Method."""

    location: Location
    header_name: str
    header_prefix: Optional[str]
    headers: Optional[Dict[str, str]]


@dataclass
class WebdriverConfig:

    """Authentication Configuration Parameters of the Webdriver Method."""

    extract_location: str
    extract_regex: str
    project: SeleniumProject
    output_format: str
    token_lifetime: Optional[int]
    extract_match_index: Optional[int]


class AuthConfigDigest(TypedDict):

    """Authentication Configuration Parameters of the Digest Method."""

    url: str
    realm: str
    nonce: str
    algorithm: AuthHashAlgorithmDigest
    domain: str
    method: HTTPMethod
    qop: Optional[str]
    nonce_count: Optional[str]
    client_nonce: Optional[str]
    opaque: Optional[str]
    headers: Optional[Dict[str, str]]


class AuthResponse(TypedDict):

    """The Processed Authentication Configuration."""

    tech: AuthTech
    headers: Dict[str, str]


Token = str


class RCFile(TypedDict):

    """RC File."""

    auth: Dict
    users: Dict


class JWTToken(TypedDict):

    """This class finds all the registered claims in the JWT token payload.

    Attributes:
        sig: Signature algorthm used in the JWT token.
        iss: Issuer of the JWT token.
        sub: Subject of the JWT token.
        aud: Audience of the JWT token -> intended for.
        exp: Expiration time of the JWT token.
        nbf: Identifies the time before which the JWT token is not yet valid.
        iat: Issued at time of the JWT token.
        jti: JWT token identifier.
        other: Other claims in the JWT token.
    """

    sig: str
    iss: Optional[str]
    sub: Optional[str]
    aud: Optional[str]
    exp: Optional[str]
    nbf: Optional[str]
    iat: Optional[str]
    jti: Optional[str]
    other: Dict[Any, Any]


# Helper Entities
AuthType = Union[AuthAWSType, AuthOAuthGrantType]



"""Aws provider."""

import sys
from enum import Enum, unique
from typing import Dict, Optional

from multiauth.entities.http import HTTPMethod, Location

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


@unique
class AuthAWSType(str, Enum):

    """The authentication flow used in the AWS authentication."""

    USER_SRP_AUTH = 'SRP'
    USER_PASSWORD_AUTH = 'Password Authentication'
    AWS_SIGNATURE = 'AWS Signature'
    REFRESH_TOKEN = 'refresh_token'


@unique
class AuthAWSChallengeResponse(str, Enum):

    """The types of challenge responses."""

    NEW_PASSWORD_REQUIRED_CHALLENGE = 'NEW_PASSWORD_REQUIRED'
    PASSWORD_VERIFIER_CHALLENGE = 'PASSWORD_VERIFIER'


@unique
class AuthHashalgorithmHawkandAWS(str, Enum):

    """The Available Hashing algorithm for Hawk authentication."""

    SHA_256 = 'sha-256'
    SHA_1 = 'sha-1'


class AuthConfigHawk(TypedDict):

    """Authentication Configuration Parameters of the Hawk Method."""

    algorithm: AuthHashalgorithmHawkandAWS
    user: Optional[str]
    nonce: Optional[str]
    ext: Optional[str]
    app: Optional[str]
    dig: Optional[str]
    timestamp: Optional[str]


class AuthConfigAWS(TypedDict):

    """Authenticaiton Configuration Parameters of the AWS Method."""

    type: AuthAWSType
    region: str
    client_id: Optional[str]
    method: Optional[HTTPMethod]
    service_name: Optional[str]
    hash_algorithm: Optional[AuthHashalgorithmHawkandAWS]
    pool_id: Optional[str]
    client_secret: Optional[str]
    location: Location
    header_name: Optional[str]
    header_prefix: Optional[str]
    headers: Optional[Dict[str, str]]



import itertools
import logging
import re
from typing import Any

logger = logging.getLogger('multiauth.providers.webdriver.extractors')


def extract_from_request_url(requests: Any, rx: str) -> list[str]:
    res = []

    for request in requests:
        if match := re.search(rx, request.url):
            res.append(match.group(1))

    return res


def extract_from_request_header(requests: Any, rx: str) -> list[str]:
    res = []

    for header in itertools.chain.from_iterable(request.headers for request in requests):
        if match := re.search(rx, header):
            res.append(match.group(1))

    return res


def extract_from_response_header(requests: Any, rx: str) -> list[str]:
    res = []
    for header in itertools.chain.from_iterable(request.response.headers for request in requests if request.response):
        if match := re.search(rx, header):
            res.append(match.group(1))

    return res


def extract_from_request_body(requests: Any, rx: str) -> list[str]:
    res = []
    for request in requests:
        if match := re.search(rx, request.body.decode()):
            res.append(match.group(1))

    return res


def extract_from_response_body(requests: Any, rx: str) -> list[str]:
    res = []
    for request in requests:
        if not request.response:
            continue
        try:
            if match := re.search(rx, request.response.body.decode()):
                res.append(match.group(1))
        except Exception as e:
            logger.debug(f'Skipping {request.url} due to error {e}')

    return res


def extract_token(location: str, rx: str, index: int | None, requests: list) -> str:
    locations = [
        'RequestURL',
        'RequestHeader',
        'RequestBody',
        'ResponseHeader',
        'ResponseBody',
    ]

    if location not in locations:
        raise ValueError(f'Invalid location `{location}`, must be one of: {locations}')

    if location == locations[0]:
        tks = extract_from_request_url(requests, rx)
    elif location == locations[1]:
        tks = extract_from_request_header(requests, rx)
    elif location == locations[2]:
        tks = extract_from_request_body(requests, rx)
    elif location == locations[3]:
        tks = extract_from_response_header(requests, rx)
    elif location == locations[4]:
        tks = extract_from_response_body(requests, rx)

    if not tks:
        raise ValueError('No token found.')

    index = index or 0
    logger.info(f'Found {len(tks)} tokens in `{location}` with regex `{rx}`. Taking index `{index}`')

    return tks[index]



"""Custom data used in utils."""

from typing import Dict, NamedTuple, Optional

from multiauth.entities.http import HTTPMethod


class Credentials(NamedTuple):

    """This is the credentials class that are the credentials found in the curl."""

    username: str
    password: str


class ParsedCurlContent(NamedTuple):

    """This is the datatype which shows the curl command."""

    method: HTTPMethod
    url: str
    data: Optional[str]
    headers: Dict[str, str]
    credentials: Optional[Credentials]



"""Utility functions independent of the library."""

import argparse
import logging
import os
import shlex
from typing import Dict, List, Mapping, Optional, TypeVar, Union
from urllib.parse import urlparse

from pydash import py_

from multiauth.entities.http import HTTPMethod
from multiauth.entities.utils import Credentials, ParsedCurlContent

Default = TypeVar('Default')
Value = TypeVar('Value')


def install_logger(logger: logging.Logger) -> None:
    """Install logger."""

    handler = logging.StreamHandler()

    formatter = os.getenv('LOG_FMT') or '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    handler.setFormatter(logging.Formatter(formatter))

    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG if os.getenv('DEBUG') else logging.INFO)

    # Ignore asyncio debug logs
    logging.getLogger('asyncio').setLevel(logging.ERROR)


def setup_logger(name: Optional[str] = None) -> logging.Logger:
    """Setup logger."""

    name = name or 'multiauth'
    logger = logging.getLogger(name)
    if not logger.hasHandlers():
        install_logger(logger)

    return logger


def dict_find_path(
    nested_dict: Mapping,
    key: str,
    prepath: str = '',
    index: Optional[int] = None,
) -> str:
    """Recursively find the path of a certain key in a Dict."""

    for k, v in nested_dict.items():
        if prepath == '':
            path = k
        elif index is not None:
            path = f'{prepath}.{index}.{k}'
        else:
            path = f'{prepath}.{k}'

        if k == key:  # found value
            return path

        if isinstance(v, Dict):
            p = dict_find_path(v, key, path, None)  # recursive call
            if p != '':
                return p

        if isinstance(v, List):
            for i, elem in enumerate(v):
                if isinstance(elem, Dict):
                    p = dict_find_path(elem, key, path, i)
                    if p != '':
                        return p

    return ''


def dict_nested_get(
    dictionary: Mapping[str, Value],
    key: str,
    default_return: Optional[Default] = None,
) -> Union[Default, Value]:
    """Search for a certain key inside a dict and returns its value (no matter the depth)"""
    return py_.get(dictionary, dict_find_path(dictionary, key, ''), default_return)


# pylint: disable=too-many-branches
def uncurl(curl: str) -> ParsedCurlContent:
    """This is a function that takes a curl as an input and analyses it."""

    parser = argparse.ArgumentParser()
    parser.add_argument('curl', help='The command curl that is analyzed in the string')
    parser.add_argument('url', help='The URL on which the curl command is working on')
    parser.add_argument(
        '-H',
        '--header',
        action='append',
        default=[],
        help='The header that are added to every request',
    )
    parser.add_argument('-X', '--request', help='the HTTP method used')
    parser.add_argument('-u', '--user', default=(), help='Specify the username password in the server authentication')
    parser.add_argument(
        '-A',
        '--user-agent',
        help='This specifies the user agent, if the value if empty than user agent will be removed from the headers',
    )
    parser.add_argument('--request-target', help='Gives the path of the target it wants to curl to')
    parser.add_argument('-e', '--referer', help='Gives the referer used in the header')
    parser.add_argument(
        '-d',
        '--data',
        '--data-binary',
        '--data-raw',
        '--data-urlencode',
        action='append',
        default=[],
        help='Data sent',
    )
    parser.add_argument('-k', '--insecure', action='store_true')

    # First we need to prepare the curl for parsing
    result_curl: List[str] = shlex.split(curl.replace('\\\n', ''))

    # Now we have to parse the curl command
    parsed_args, _ = parser.parse_known_args(result_curl)

    # Now we have to analyze what we have
    # First find out what type of method are we using
    method: HTTPMethod = 'GET'
    if parsed_args.request:
        method = parsed_args.request.upper()
    else:
        if parsed_args.data:
            method = 'POST'
        else:
            method = 'GET'

    # Now we have to extract the headers
    headers: Dict[str, str] = {}
    for header in parsed_args.header:
        header_prefix, header_value = header.split(':', 1)
        headers[header_prefix] = header_value.strip()

    if parsed_args.user_agent:
        headers['User-Agent'] = parsed_args.user_agent

    if parsed_args.referer:
        headers['Referer'] = parsed_args.user_agent

    datas: list = parsed_args.data
    final_data: Optional[str] = None
    if len(datas) != 0:
        final_data = datas[0]
    if len(datas) != 1:
        for data in datas[1:]:
            final_data += '&' + data

    url: str = ''
    if parsed_args.request_target:
        parsed_url = urlparse(parsed_args.url)
        if parsed_args.request_target[-1] == '/':
            url = parsed_url.scheme + '://' + parsed_url.netloc + parsed_args.request_target
        else:
            url = parsed_url.scheme + '://' + parsed_url.netloc + '/' + parsed_args.request_target
    else:
        url = parsed_args.url

    # Add auth
    credentials: Optional[Credentials] = None
    if parsed_args.user:
        user_name, password = parsed_args.user.split(':', 1)
        credentials = Credentials(user_name, password)

    return ParsedCurlContent(method, url, final_data, headers, credentials)



"""Implementation of the Manual authentication schema."""

from typing import Any, Dict, cast

from multiauth.entities.errors import AuthenticationError
from multiauth.entities.main import AuthResponse, AuthTech
from multiauth.manager import User


def manual_authenticator(user: User) -> AuthResponse:
    """This function implements the Manual Authentication Schema.

    It simply take the headers that are found in the user credentials and puts them in the AuthResponse.
    """

    auth_response = AuthResponse(
        {
            'headers': {},
            'tech': AuthTech.MANUAL,
        },
    )

    headers = user.headers
    if not headers:
        if not user.credentials:
            raise AuthenticationError('Configuration file error. Missing credentials')
        if 'headers' not in user.credentials:
            raise AuthenticationError('Please input the necessary authentication headers.')

        headers = user.credentials['headers']

    auth_response['headers'] = cast(Dict[str, Any], headers)
    auth_response['tech'] = AuthTech.MANUAL

    return auth_response



"""Multiauth errors."""


class AuthenticationError(Exception):

    """Authentication error."""

    message = 'Your authentication failed.'


class ExpiredTokenError(Exception):

    """Token expired error."""

    message = 'Your token has expired.'


class InvalidConfigurationError(Exception):

    """Generic excpetion raised for escaperc misconfiguration."""

    errno = 2

    message: str
    path: str

    def __init__(self, message: str, path: str) -> None:
        """Create the exception."""

        super().__init__(message)

        self.message = message
        self.path = path
